# Lint as: python3
"""Helper class to load and manage the robot URDF file in simulation."""

import collections
from typing import Dict, Text, Tuple

import gin
import numpy as np

from pybullet_utils import bullet_client

# Base link does not have a parent joint. So we just use the string "robot_base"
# for reference. The corresponding link/joint id is always -1 in pybullet.
ROBOT_BASE = "robot_base"


def _sub_dict(joint_name_to_id: Dict[Text, int],
              joint_names: Tuple[Text]) -> Dict[Text, int]:
  sub_dict = collections.OrderedDict()
  if joint_names is None:
    return sub_dict
  for name in joint_names:
    sub_dict[name] = joint_name_to_id[name]
  return sub_dict


def convert_to_urdf_joint_angles(
    robot_space_joint_angles: np.ndarray,
    joint_offsets: np.ndarray,
    joint_directions: np.ndarray,
):
  return robot_space_joint_angles * joint_directions + joint_offsets


def convert_to_robot_joint_angles(
    urdf_space_joint_angles: np.ndarray,
    joint_offsets: np.ndarray,
    joint_directions: np.ndarray,
):
  return (urdf_space_joint_angles - joint_offsets) * joint_directions


@gin.configurable
class RobotUrdfLoader(object):
  """A abstract class to manage the robot urdf in sim."""

  def __init__(
      self,
      pybullet_client: bullet_client.BulletClient,
      urdf_path: Text,
      constrained_base: bool = False,
      enable_self_collision: bool = True,
      init_base_position: Tuple[float] = (0, 0, 0),
      init_base_orientation_quaternion: Tuple[float] = None,
      init_base_orientation_rpy: Tuple[float] = None,
      base_names: Tuple[Text] = None,
      init_joint_angles: Dict[Text, float] = None,
      joint_offsets: Dict[Text, float] = None,
      joint_directions: Dict[Text, int] = None,
      motor_names: Tuple[Text] = None,
      end_effector_names: Tuple[Text] = None,
      user_group: Dict[Text, Tuple[Text]] = None,
  ):
    """Initialize the class.

    Args:
      pybullet_client: A pybullet client.
      urdf_path: The path to the URDF to load.
      constrained_base: Whether to create a FIXED constraint to the base of the
        URDF. Needs to be True for kinematic robots. This allows us to "hang"
        the simulated robot in air, and the hanging point can follow arbitrary
        provided paths.
      enable_self_collision: Determines if the robot can collide with itself.
      init_base_position: The base x, y, z after loading.
      init_base_orientation_quaternion: The base rotation after loading.
      init_base_orientation_rpy: The base rotation after loading, presented in
        roll, pitch, yaw.
      base_names: The base joint names. Used to find additional links that
        belong to the base. Optional because the base might only contain a
        single mesh/block, which always has the id of "-1".
      init_joint_angles: Maps joint name to the desired joint pose after loading
        URDF. If not provided, will use the URDF default. This can be a subset
        of all joints in the URDF. This should be in the robot joint convention,
        which can be different from the URDF convention.
      joint_offsets: The "zero" position of joint angles in the URDF space.
      joint_directions: To convert between robot sdk/control and urdf joint
        convention.
      motor_names: The motor joint names in the URDF. Typically a subset of all
        movable joints/DoFs.
      end_effector_names: A subset of joints specifying the end-effector
        joint(s). For example for legged robots the end effectors are the toe
        joints (if provided). For arms this group includes left and right
        grippers.
      user_group: User defined joint groups. For example for quadrupeds, we may
        want to organize all joints according to which leg they belong to.
    """
    self._pybullet_client = pybullet_client
    self._urdf_path = urdf_path
    self._init_base_position = init_base_position
    if init_base_orientation_quaternion is not None:
      self._init_base_orientation_quaternion = init_base_orientation_quaternion
    else:
      if init_base_orientation_rpy is None:
        raise ValueError("Either init_base_orientation_quaterion "
                         "or init_base_orientation_rpy is required")
      self._init_base_orientation_quaternion = (
          self._pybullet_client.getQuaternionFromEuler(
              init_base_orientation_rpy))
    self._constrained_base = constrained_base
    self._enable_self_collision = enable_self_collision
    self._base_names = base_names
    self._init_joint_angles = init_joint_angles
    self._joint_offsets = joint_offsets
    self._joint_directions = joint_directions
    self._motor_names = motor_names
    self._end_effector_names = end_effector_names
    self._user_group = user_group
    self.load(
        enable_self_collision=enable_self_collision,
        init_base_position=init_base_position,
        init_base_orientation_quaternion=init_base_orientation_quaternion,
        init_joint_angles=init_joint_angles)

  def get_base_id_dict(self, name: Tuple[Text] = None) -> Dict[Text, int]:
    if name is None:
      return self._base_dict
    return _sub_dict(self._base_dict, name)

  def get_joint_id_dict(self, name: Tuple[Text] = None) -> Dict[Text, int]:
    if name is None:
      return self._joint_name_to_id
    return _sub_dict(self._joint_name_to_id, name)

  def get_link_id_dict(self, name: Tuple[Text] = None) -> Dict[Text, int]:
    if name is None:
      return self._link_name_to_id
    return _sub_dict(self._link_name_to_id, name)

  def get_motor_id_dict(self, name: Tuple[Text] = None) -> Dict[Text, int]:
    if name is None:
      return self._motor_dict
    return _sub_dict(self._motor_dict, name)

  def get_joint_direction_dict(self,
                               name: Tuple[Text] = None) -> Dict[Text, int]:
    if name is None:
      return self._joint_directions
    return _sub_dict(self._joint_directions, name)

  def get_joint_offset_dict(self,
                            name: Tuple[Text] = None) -> Dict[Text, float]:
    if name is None:
      return self._joint_offsets
    return _sub_dict(self._joint_offsets, name)

  def get_end_effector_id_dict(self,
                               name: Tuple[Text] = None) -> Dict[Text, int]:
    if name is None:
      return self._end_effector_dict
    return _sub_dict(self._end_effector_dict, name)

  @property
  def robot_id(self):
    """Returns the unique object instance id of this loaded URDF in pybullet.

    Note this is different from all other get_{}_id APIs, which returns the
    joint/link id within this robot instance.

    Returns:
      The object id as returned by loadURDF.
    """
    return self._robot_id

  @property
  def all_joint_info(self):
    return self._all_joint_info

  @property
  def user_dict(self):
    return self._user_dict

  @property
  def motor_names(self):
    return self._motor_names

  @property
  def constrained_base(self):
    return self._constrained_base

  def _build_base_dict(self):
    """Builds the base joints dictionary.

    In pybullet, a link's id within the robot always equal to its parent joint
    id. So this base joint dict functionaly is equivalent to the base link dict.
    The dictionary may only contain {ROBOT_BASE: -1} if self._base_names is
    empty.

    Returns:
      The base link (joint) ordered dictionary.
    """
    base_dict = collections.OrderedDict()
    if self._base_names is None:
      base_dict[ROBOT_BASE] = -1
    else:
      base_dict.update(_sub_dict(self._joint_name_to_id, self._base_names))
    return base_dict

  def _build_user_dict(self):
    """Builds a dictionary using user defined joint groups."""
    user_dict = collections.OrderedDict()
    if self._user_group is None:
      return user_dict
    for group_name, group_joint_names in self._user_group.items():
      user_dict[group_name] = collections.OrderedDict()
      user_dict[group_name].update(
          _sub_dict(self._joint_name_to_id, group_joint_names))
    return user_dict

  def _build_all_joint_dict(self):
    """Extracts all joints from the URDF.

    Finds all joints (fixed or movable) in the URDF and extracts the info. This
    includes actuated joints (i.e. motors), and non-actuated joints, e.g. the
    passive joints in Minitaur's four bar mechanism, and fixed joints connecting
    the toe and the lower legs, etc.

    Returns:
      number of joints, all joint information as returned by pybullet, and the
      joint_name_to_id dictionary.

    """
    num_joints = self._pybullet_client.getNumJoints(self._robot_id)
    all_joint_info = [
        self._pybullet_client.getJointInfo(self._robot_id, i)
        for i in range(num_joints)
    ]

    # Remove the default joint dampings to increase sim fidelity.
    for joint_info in all_joint_info:
      joint_id = joint_info[0]
      self._pybullet_client.changeDynamics(
          joint_id, -1, linearDamping=0, angularDamping=0)

    joint_name_to_id = collections.OrderedDict()
    link_name_to_id = collections.OrderedDict([(ROBOT_BASE, -1)])
    for joint_info in all_joint_info:
      joint_name = joint_info[1].decode("UTF-8")
      joint_id = joint_info[0]
      joint_name_to_id[joint_name] = joint_id
      # Index 12 is the name of the joint's child link, and in PyBullet a child
      # link id is always equal to its parent joint id.
      link_name_to_id[joint_info[12].decode("UTF-8")] = joint_id

    return num_joints, all_joint_info, joint_name_to_id, link_name_to_id

  def load(
      self,
      enable_self_collision: bool = None,
      init_base_position: Tuple[float] = None,
      init_base_orientation_quaternion: Tuple[float] = None,
      init_joint_angles: Dict[Text, float] = None,
  ):
    """Reloads the URDF and rebuilds the dictionaries."""
    if enable_self_collision is None:
      enable_self_collision = self._enable_self_collision
    if init_base_position is None:
      init_base_position = self._init_base_position
    if init_base_orientation_quaternion is None:
      init_base_orientation_quaternion = self._init_base_orientation_quaternion

    self._robot_id = self._load_urdf(enable_self_collision, init_base_position,
                                     init_base_orientation_quaternion)

    (self._num_joints, self._all_joint_info, self._joint_name_to_id,
     self._link_name_to_id) = self._build_all_joint_dict()
    self._base_dict = self._build_base_dict()
    self._motor_dict = _sub_dict(self._joint_name_to_id, self._motor_names)
    self._end_effector_dict = _sub_dict(self._joint_name_to_id,
                                        self._end_effector_names)
    self._user_dict = self._build_user_dict()

    self.reset_base_pose(init_base_position, init_base_orientation_quaternion)
    self.reset_joint_angles(init_joint_angles)

  def _load_urdf(self, enable_self_collision: bool,
                 init_base_position: Tuple[float],
                 init_base_orientation_quaternion: Tuple[float]) -> int:
    """Loads the URDF and returns the pybullet id."""
    try:
      if enable_self_collision:
        return self._pybullet_client.loadURDF(
            self._urdf_path,
            init_base_position,
            init_base_orientation_quaternion,
            useFixedBase=self._constrained_base,
            flags=self._pybullet_client.URDF_USE_SELF_COLLISION)
      else:
        return self._pybullet_client.loadURDF(
            self._urdf_path,
            init_base_position,
            init_base_orientation_quaternion,
            useFixedBase=self._constrained_base,
        )
    except:
      print("!!!!!!!!!!!!!!!!")
      print("Error: cannot find file:")
      print(self._urdf_path)
      import sys
      sys.exit(0)
      
  def reset_joint_angles(self, joint_angles: Dict[Text, float] = None):
    """Resets the joint angles.

    Resets the joint poses. This is instanteneously and will ignore the physics
    (e.g. collisions, inertias, etc). Should only be used during the episode
    reset time. Does not work for real robots (other than changing the
    visualization). This API has no effect if both the input joint_angles and
    the self._init_joint_angles are None.

    Args:
      joint_angles: The joint angles in the robot's joint space.

    Raises:
      AttributeError if the joint directions and joint offsets are not provided
      during init.
    """
    if self._init_joint_angles is None:
      return

    if joint_angles is None:
      joint_angles = self._init_joint_angles

    if self._joint_directions is None or self._joint_offsets is None:
      raise AttributeError(
          "joint directions and joint offsets not provided in __init__")

    for joint_name, angle in joint_angles.items():
      urdf_joint_angle = angle * self._joint_directions[
          joint_name] + self._joint_offsets[joint_name]
      self._pybullet_client.resetJointState(
          self._robot_id,
          self._joint_name_to_id[joint_name],
          urdf_joint_angle,
          targetVelocity=0)

  def reset_base_pose(
      self,
      base_position: Tuple[float] = None,
      base_orientation_quaternion: Tuple[float] = None,
  ):
    """Resets the base position and orientation.

    Instanteneously re-position the base pose of the robot. Does not work for
    real robots except for the visualization.

    Args:
      base_position: Base x, y, z position.
      base_orientation_quaternion: Base rotation.
    """
    if base_position is None:
      base_position = self._init_base_position
    if base_orientation_quaternion is None:
      base_orientation_quaternion = self._init_base_orientation_quaternion
    self._pybullet_client.resetBasePositionAndOrientation(
        self._robot_id, base_position, base_orientation_quaternion)
    self._pybullet_client.resetBaseVelocity(self._robot_id, (0, 0, 0),
                                            (0, 0, 0))

  @property
  def com_dof(self):
    return 0 if self._constrained_base else 6



# Lint as: python3
"""The on robot sensor classes."""

from typing import Any, Callable, Iterable, Optional, Sequence, Type, Text, Tuple, Union

import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor

from pybullet_envs.minitaur.envs_v2.utilities import noise_generators


def _convert_to_np_array(inputs: Union[float, Tuple[float], np.ndarray], dim):
  """Converts the inputs to a numpy array.

  Args:
    inputs: The input scalar or array.
    dim: The dimension of the converted numpy array.

  Returns:
    The converted numpy array.

  Raises:
    ValueError: If the inputs is an array whose dimension does not match the
    provided dimension.
  """
  outputs = None
  if isinstance(inputs, (tuple, np.ndarray)):
    outputs = np.array(inputs)
  else:
    outputs = np.full(dim, inputs)

  if len(outputs) != dim:
    raise ValueError("The inputs array has a different dimension {}"
                     " than provided, which is {}.".format(len(outputs), dim))

  return outputs


@gin.configurable
class MotorAngleSensor(sensor.Sensor):
  """A sensor that reads motor angles from the robot."""

  def __init__(self,
               name: Text = "MotorAngle",
               dtype: Type[Any] = np.float64,
               lower_bound: Optional[Union[float, Iterable[float]]] = None,
               upper_bound: Optional[Union[float, Iterable[float]]] = None,
               noise_generator: Union[Callable[..., Any],
                                      noise_generators.NoiseGenerator] = None,
               sensor_latency: Union[float, Sequence[float]] = 0.0,
               observe_sine_cosine: bool = False):
    """Initializes the class.

    Args:
      name: The name of the sensor.
      dtype: The datatype of this sensor.
      lower_bound: The optional lower bounds of the sensor reading. If not
        provided, will extract from the motor limits of the robot class.
      upper_bound: The optional upper bounds of the sensor reading. If not
        provided, will extract from the motor limits of the robot class.
      noise_generator: Adds noise to the sensor readings.
      sensor_latency: There are two ways to use this expected sensor latency.
        For both methods, the latency should be in the same unit as the sensor
        data timestamp. 1. As a single float number, the observation will be a
        1D array. For real robots, this should be set to 0.0. 2. As a array of
        floats, the observation will be a 2D array based on how long the history
        need to be. Thus, [0.0, 0.1, 0.2] is a history length of 3.
       observe_sine_cosine: whether to observe motor angles as sine and cosine
         values.
    """
    super().__init__(
        name=name,
        sensor_latency=sensor_latency,
        interpolator_fn=sensor.linear_obs_blender)
    self._noise_generator = noise_generator
    self._dtype = dtype
    self._lower_bound = lower_bound
    self._upper_bound = upper_bound
    self._observe_sine_cosine = observe_sine_cosine

  def set_robot(self, robot):
    self._robot = robot
    # Creates the observation space based on the robot motor limitations.
    if self._observe_sine_cosine:
      lower_bound = _convert_to_np_array(-1, 2 * self._robot.num_motors)
    elif self._lower_bound:
      lower_bound = _convert_to_np_array(self._lower_bound,
                                         self._robot.num_motors)
    else:
      lower_bound = _convert_to_np_array(
          self._robot.motor_limits.angle_lower_limits, self._robot.num_motors)
    if self._observe_sine_cosine:
      upper_bound = _convert_to_np_array(1, 2 * self._robot.num_motors)
    elif self._upper_bound:
      upper_bound = _convert_to_np_array(self._upper_bound,
                                         self._robot.num_motors)
    else:
      upper_bound = _convert_to_np_array(
          self._robot.motor_limits.angle_upper_limits, self._robot.num_motors)

    self._observation_space = self._stack_space(
        gym.spaces.Box(low=lower_bound, high=upper_bound, dtype=self._dtype))

  def _get_original_observation(self):
    if self._observe_sine_cosine:
      return self._robot.timestamp, np.hstack(
          (np.cos(self._robot.motor_angles), np.sin(self._robot.motor_angles)))
    else:
      return self._robot.timestamp, self._robot.motor_angles

  def get_observation(self) -> np.ndarray:
    delayed_observation = super().get_observation()
    if self._noise_generator:
      if callable(self._noise_generator):
        return self._noise_generator(delayed_observation)
      else:
        return self._noise_generator.add_noise(delayed_observation)

    return delayed_observation



"""Pybullet simulation of a vision60 robot."""
import math
import os

import gin
import numpy as np

from pybullet_envs.minitaur.robots import laikago_motor
from pybullet_envs.minitaur.robots import minitaur
from pybullet_envs.minitaur.robots import robot_config

NUM_MOTORS = 12
NUM_LEGS = 4
MOTOR_NAMES = [
    "torso_to_abduct_fl_j",  # Left front abduction (hip0).
    "abduct_fl_to_thigh_fl_j",  # Left front hip (upper0).
    "thigh_fl_to_knee_fl_j",  # Left front knee (lower0).
    "torso_to_abduct_hl_j",  # Left rear abduction (hip1).
    "abduct_hl_to_thigh_hl_j",  # Left rear hip (upper1).
    "thigh_hl_to_knee_hl_j",  # Left rear knee (lower1).
    "torso_to_abduct_fr_j",  # Right front abduction (hip2).
    "abduct_fr_to_thigh_fr_j",  # Right front hip (upper2).
    "thigh_fr_to_knee_fr_j",  # Right front knee (lower2).
    "torso_to_abduct_hr_j",  # Right rear abduction (hip3).
    "abduct_hr_to_thigh_hr_j",  # Right rear hip (upper3).
    "thigh_hr_to_knee_hr_j",  # Right rear knee (lower3).
]
_DEFAULT_TORQUE_LIMITS = [12, 18, 12] * 4
INIT_RACK_POSITION = [0, 0, 1.4]
INIT_POSITION = [0, 0, 0.4]
JOINT_DIRECTIONS = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
HIP_JOINT_OFFSET = 0.0
UPPER_LEG_JOINT_OFFSET = 0.0
KNEE_JOINT_OFFSET = 0.0
DOFS_PER_LEG = 3
JOINT_OFFSETS = np.array(
    [HIP_JOINT_OFFSET, UPPER_LEG_JOINT_OFFSET, KNEE_JOINT_OFFSET] * 4)
PI = math.pi
DEFAULT_ABDUCTION_ANGLE = 0.0
DEFAULT_HIP_ANGLE = -1.1
DEFAULT_KNEE_ANGLE = 2.3
# Bases on the readings from 's default pose.
INIT_MOTOR_ANGLES = [
    DEFAULT_ABDUCTION_ANGLE, DEFAULT_HIP_ANGLE, DEFAULT_KNEE_ANGLE
] * NUM_LEGS
DEFAULT_LOCAL_TOE_POSITIONS = [[0.17, -0.11, -0.16], [0.17, 0.11, -0.16],
                               [-0.20, -0.11, -0.16], [-0.20, 0.11, -0.16]]


@gin.configurable
class MiniCheetah(minitaur.Minitaur):
  """A simulation for the mini cheetah robot."""

  def __init__(self, **kwargs):
    if "motor_kp" not in kwargs:
      kwargs["motor_kp"] = 100.0
    if "motor_kd" not in kwargs:
      kwargs["motor_kd"] = 2.0
    if "motor_torque_limits" not in kwargs:
      kwargs["motor_torque_limits"] = _DEFAULT_TORQUE_LIMITS

    # The follwing parameters are fixed for the vision60 robot.
    kwargs["num_motors"] = NUM_MOTORS
    kwargs["dofs_per_leg"] = DOFS_PER_LEG
    kwargs["motor_direction"] = JOINT_DIRECTIONS
    kwargs["motor_offset"] = JOINT_OFFSETS
    kwargs["motor_overheat_protection"] = False
    kwargs["motor_model_class"] = laikago_motor.LaikagoMotorModel
    super(MiniCheetah, self).__init__(**kwargs)

  def _LoadRobotURDF(self):
    mini_cheetah_urdf_path = "mini_cheetah/mini_cheetah.urdf"
    if self._self_collision_enabled:
      self.quadruped = self._pybullet_client.loadURDF(
          mini_cheetah_urdf_path,
          self._GetDefaultInitPosition(),
          self._GetDefaultInitOrientation(),
          flags=self._pybullet_client.URDF_USE_SELF_COLLISION)
    else:
      self.quadruped = self._pybullet_client.loadURDF(
          mini_cheetah_urdf_path, self._GetDefaultInitPosition(),
          self._GetDefaultInitOrientation())

  def _SettleDownForReset(self, default_motor_angles, reset_time):
    self.ReceiveObservation()
    for _ in range(500):
      self.ApplyAction(
          INIT_MOTOR_ANGLES,
          motor_control_mode=robot_config.MotorControlMode.POSITION)
      self._pybullet_client.stepSimulation()
      self.ReceiveObservation()
    if default_motor_angles is not None:
      num_steps_to_reset = int(reset_time / self.time_step)
      for _ in range(num_steps_to_reset):
        self.ApplyAction(
            default_motor_angles,
            motor_control_mode=robot_config.MotorControlMode.POSITION)
        self._pybullet_client.stepSimulation()
        self.ReceiveObservation()

  def GetURDFFile(self):
    return os.path.join(self._urdf_root, "mini_cheetah/mini_cheetah.urdf")

  def ResetPose(self, add_constraint):
    del add_constraint
    for name in self._joint_name_to_id:
      joint_id = self._joint_name_to_id[name]
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=(joint_id),
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=0)
    for name, i in zip(MOTOR_NAMES, range(len(MOTOR_NAMES))):
      angle = INIT_MOTOR_ANGLES[i]
      self._pybullet_client.resetJointState(
          self.quadruped, self._joint_name_to_id[name], angle, targetVelocity=0)

  def _BuildUrdfIds(self):
    pass

  def _GetMotorNames(self):
    return MOTOR_NAMES

  def _GetDefaultInitPosition(self):
    if self._on_rack:
      return INIT_RACK_POSITION
    else:
      return INIT_POSITION

  def _GetDefaultInitOrientation(self):
    init_orientation = [0, 0, 0, 1.0]
    return init_orientation



import argparse
import numpy

from pybullet_envs.minitaur.envs import minitaur_logging

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--log_file', help='path to protobuf file', default='')
args = parser.parse_args()
logging = minitaur_logging.MinitaurLogging()
episode = logging.restore_episode(args.log_file)
#print(dir (episode))
#print("episode=",episode)
fields = episode.ListFields()

recs = []

for rec in fields[0][1]:
  #print(rec.time)
  for motorState in rec.motor_states:
    #print("motorState.angle=",motorState.angle)
    #print("motorState.velocity=",motorState.velocity)
    #print("motorState.action=",motorState.action)
    #print("motorState.torque=",motorState.torque)
    recs.append([motorState.angle, motorState.velocity, motorState.action, motorState.torque])

a = numpy.array(recs)
numpy.savetxt("motorq_qdot_action_torque.csv", a, delimiter=",")



import numpy as np

RAD_TO_DEG = 57.2957795
DEG_TO_RAD = 1.0 / RAD_TO_DEG
INVALID_IDX = -1


def lerp(x, y, t):
  return (1 - t) * x + t * y


def log_lerp(x, y, t):
  return np.exp(lerp(np.log(x), np.log(y), t))


def flatten(arr_list):
  return np.concatenate([np.reshape(a, [-1]) for a in arr_list], axis=0)


def flip_coin(p):
  rand_num = np.random.binomial(1, p, 1)
  return rand_num[0] == 1



import gym, gym.spaces, gym.utils, gym.utils.seeding
import numpy as np
import pybullet
import os

from pybullet_utils import bullet_client

from pkg_resources import parse_version

try:
  if os.environ["PYBULLET_EGL"]:
    import pkgutil
except:
  pass

class MJCFBaseBulletEnv(gym.Env):
  """
	Base class for Bullet physics simulation loading MJCF (MuJoCo .xml) environments in a Scene.
	These environments create single-player scenes and behave like normal Gym environments, if
	you don't use multiplayer.
	"""

  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 60}

  def __init__(self, robot, render=False):
    self.scene = None
    self.physicsClientId = -1
    self.ownsPhysicsClient = 0
    self.camera = Camera(self)
    self.isRender = render
    self.robot = robot
    self.seed()
    self._cam_dist = 3
    self._cam_yaw = 0
    self._cam_pitch = -30
    self._render_width = 320
    self._render_height = 240

    self.action_space = robot.action_space
    self.observation_space = robot.observation_space
    #self.reset()

  def configure(self, args):
    self.robot.args = args

  def seed(self, seed=None):
    self.np_random, seed = gym.utils.seeding.np_random(seed)
    self.robot.np_random = self.np_random  # use the same np_randomizer for robot as for env
    return [seed]

  def reset(self):
    if (self.physicsClientId < 0):
      self.ownsPhysicsClient = True

      if self.isRender:
        self._p = bullet_client.BulletClient(connection_mode=pybullet.GUI)
      else:
        self._p = bullet_client.BulletClient()
      self._p.resetSimulation()
      self._p.setPhysicsEngineParameter(deterministicOverlappingPairs=1)
      #optionally enable EGL for faster headless rendering
      try:
        if os.environ["PYBULLET_EGL"]:
          con_mode = self._p.getConnectionInfo()['connectionMethod']
          if con_mode==self._p.DIRECT:
            egl = pkgutil.get_loader('eglRenderer')
            if (egl):
              self._p.loadPlugin(egl.get_filename(), "_eglRendererPlugin")
            else:
              self._p.loadPlugin("eglRendererPlugin")
      except:
        pass
      self.physicsClientId = self._p._client
      self._p.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)

    if self.scene is None:
      self.scene = self.create_single_player_scene(self._p)
    if not self.scene.multiplayer and self.ownsPhysicsClient:
      self.scene.episode_restart(self._p)

    self.robot.scene = self.scene

    self.frame = 0
    self.done = 0
    self.reward = 0
    dump = 0
    s = self.robot.reset(self._p)
    self.potential = self.robot.calc_potential()
    return s

  def camera_adjust(self):
    pass

  def render(self, mode='human', close=False):
  
    if mode == "human":
      self.isRender = True
    if self.physicsClientId>=0:
      self.camera_adjust()

    if mode != "rgb_array":
      return np.array([])

    base_pos = [0, 0, 0]
    if (hasattr(self, 'robot')):
      if (hasattr(self.robot, 'body_real_xyz')):
        base_pos = self.robot.body_real_xyz
    if (self.physicsClientId>=0):
      view_matrix = self._p.computeViewMatrixFromYawPitchRoll(cameraTargetPosition=base_pos,
                                                            distance=self._cam_dist,
                                                            yaw=self._cam_yaw,
                                                            pitch=self._cam_pitch,
                                                            roll=0,
                                                            upAxisIndex=2)
      proj_matrix = self._p.computeProjectionMatrixFOV(fov=60,
                                                     aspect=float(self._render_width) /
                                                     self._render_height,
                                                     nearVal=0.1,
                                                     farVal=100.0)
      (_, _, px, _, _) = self._p.getCameraImage(width=self._render_width,
                                              height=self._render_height,
                                              viewMatrix=view_matrix,
                                              projectionMatrix=proj_matrix,
                                              renderer=pybullet.ER_BULLET_HARDWARE_OPENGL)

      self._p.configureDebugVisualizer(self._p.COV_ENABLE_SINGLE_STEP_RENDERING,1)
    else:
      px = np.array([[[255,255,255,255]]*self._render_width]*self._render_height, dtype=np.uint8)
    rgb_array = np.array(px, dtype=np.uint8)
    rgb_array = np.reshape(np.array(px), (self._render_height, self._render_width, -1))
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def close(self):
    if (self.ownsPhysicsClient):
      if (self.physicsClientId >= 0):
        self._p.disconnect()
    self.physicsClientId = -1

  def HUD(self, state, a, done):
    pass

  # def step(self, *args, **kwargs):
  # 	if self.isRender:
  # 		base_pos=[0,0,0]
  # 		if (hasattr(self,'robot')):
  # 			if (hasattr(self.robot,'body_xyz')):
  # 				base_pos = self.robot.body_xyz
  # 				# Keep the previous orientation of the camera set by the user.
  # 				#[yaw, pitch, dist] = self._p.getDebugVisualizerCamera()[8:11]
  # 				self._p.resetDebugVisualizerCamera(3,0,0, base_pos)
  #
  #
  # 	return self.step(*args, **kwargs)
  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed


class Camera:

  def __init__(self, env):
    self.env = env
    pass

  def move_and_look_at(self, i, j, k, x, y, z):
    lookat = [x, y, z]
    camInfo = self.env._p.getDebugVisualizerCamera()
    
    distance = camInfo[10]
    pitch = camInfo[9]
    yaw = camInfo[8]
    self.env._p.resetDebugVisualizerCamera(distance, yaw, pitch, lookat)



from . import enjoy_kuka_grasping
from . import enjoy_pybullet_cartpole
from . import enjoy_pybullet_racecar
from . import enjoy_pybullet_zed_racecar
from . import train_kuka_cam_grasping
from . import train_kuka_grasping
from . import train_pybullet_cartpole
from . import train_pybullet_racecar
from . import train_pybullet_zed_racecar



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for OpenAI Gym environments."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import atexit
import multiprocessing
import sys
import traceback

import gym
import gym.spaces
import numpy as np
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf


class AutoReset(object):
  """Automatically reset environment when the episode is done."""

  def __init__(self, env):
    self._env = env
    self._done = True

  def __getattr__(self, name):
    return getattr(self._env, name)

  def step(self, action):
    if self._done:
      observ, reward, done, info = self._env.reset(), 0.0, False, {}
    else:
      observ, reward, done, info = self._env.step(action)
    self._done = done
    return observ, reward, done, info

  def reset(self):
    self._done = False
    return self._env.reset()


class ActionRepeat(object):
  """Repeat the agent action multiple steps."""

  def __init__(self, env, amount):
    self._env = env
    self._amount = amount

  def __getattr__(self, name):
    return getattr(self._env, name)

  def step(self, action):
    done = False
    total_reward = 0
    current_step = 0
    while current_step < self._amount and not done:
      observ, reward, done, info = self._env.step(action)
      total_reward += reward
      current_step += 1
    return observ, total_reward, done, info


class RandomStart(object):
  """Perform random number of random actions at the start of the episode."""

  def __init__(self, env, max_steps):
    self._env = env
    self._max_steps = max_steps

  def __getattr__(self, name):
    return getattr(self._env, name)

  def reset(self):
    observ = self._env.reset()
    random_steps = np.random.randint(0, self._max_steps)
    for _ in range(random_steps):
      action = self._env.action_space.sample()
      observ, unused_reward, done, unused_info = self._env.step(action)
      if done:
        tf.logging.warning('Episode ended during random start.')
        return self.reset()
    return observ


class FrameHistory(object):
  """Augment the observation with past observations."""

  def __init__(self, env, past_indices, flatten):
    """Augment the observation with past observations.

    Implemented as a Numpy ring buffer holding the necessary past observations.

    Args:
      env: OpenAI Gym environment to wrap.
      past_indices: List of non-negative integers indicating the time offsets
        from the current time step of observations to include.
      flatten: Concatenate the past observations rather than stacking them.

    Raises:
      KeyError: The current observation is not included in the indices.
    """
    if 0 not in past_indices:
      raise KeyError('Past indices should include 0 for the current frame.')
    self._env = env
    self._past_indices = past_indices
    self._step = 0
    self._buffer = None
    self._capacity = max(past_indices)
    self._flatten = flatten

  def __getattr__(self, name):
    return getattr(self._env, name)

  @property
  def observation_space(self):
    low = self._env.observation_space.low
    high = self._env.observation_space.high
    low = np.repeat(low[None, ...], len(self._past_indices), 0)
    high = np.repeat(high[None, ...], len(self._past_indices), 0)
    if self._flatten:
      low = np.reshape(low, (-1,) + low.shape[2:])
      high = np.reshape(high, (-1,) + high.shape[2:])
    return gym.spaces.Box(low, high)

  def step(self, action):
    observ, reward, done, info = self._env.step(action)
    self._step += 1
    self._buffer[self._step % self._capacity] = observ
    observ = self._select_frames()
    return observ, reward, done, info

  def reset(self):
    observ = self._env.reset()
    self._buffer = np.repeat(observ[None, ...], self._capacity, 0)
    self._step = 0
    return self._select_frames()

  def _select_frames(self):
    indices = [(self._step - index) % self._capacity for index in self._past_indices]
    observ = self._buffer[indices]
    if self._flatten:
      observ = np.reshape(observ, (-1,) + observ.shape[2:])
    return observ


class FrameDelta(object):
  """Convert the observation to a difference from the previous observation."""

  def __init__(self, env):
    self._env = env
    self._last = None

  def __getattr__(self, name):
    return getattr(self._env, name)

  @property
  def observation_space(self):
    low = self._env.observation_space.low
    high = self._env.observation_space.high
    low, high = low - high, high - low
    return gym.spaces.Box(low, high)

  def step(self, action):
    observ, reward, done, info = self._env.step(action)
    delta = observ - self._last
    self._last = observ
    return delta, reward, done, info

  def reset(self):
    observ = self._env.reset()
    self._last = observ
    return observ


class RangeNormalize(object):
  """Normalize the specialized observation and action ranges to [-1, 1]."""

  def __init__(self, env, observ=None, action=None):
    self._env = env
    self._should_normalize_observ = (observ is not False and
                                     self._is_finite(self._env.observation_space))
    if observ is True and not self._should_normalize_observ:
      raise ValueError('Cannot normalize infinite observation range.')
    if observ is None and not self._should_normalize_observ:
      tf.logging.info('Not normalizing infinite observation range.')
    self._should_normalize_action = (action is not False and
                                     self._is_finite(self._env.action_space))
    if action is True and not self._should_normalize_action:
      raise ValueError('Cannot normalize infinite action range.')
    if action is None and not self._should_normalize_action:
      tf.logging.info('Not normalizing infinite action range.')

  def __getattr__(self, name):
    return getattr(self._env, name)

  @property
  def observation_space(self):
    space = self._env.observation_space
    if not self._should_normalize_observ:
      return space
    return gym.spaces.Box(-np.ones(space.shape), np.ones(space.shape))

  @property
  def action_space(self):
    space = self._env.action_space
    if not self._should_normalize_action:
      return space
    return gym.spaces.Box(-np.ones(space.shape), np.ones(space.shape))

  def step(self, action):
    if self._should_normalize_action:
      action = self._denormalize_action(action)
    observ, reward, done, info = self._env.step(action)
    if self._should_normalize_observ:
      observ = self._normalize_observ(observ)
    return observ, reward, done, info

  def reset(self):
    observ = self._env.reset()
    if self._should_normalize_observ:
      observ = self._normalize_observ(observ)
    return observ

  def _denormalize_action(self, action):
    min_ = self._env.action_space.low
    max_ = self._env.action_space.high
    action = (action + 1) / 2 * (max_ - min_) + min_
    return action

  def _normalize_observ(self, observ):
    min_ = self._env.observation_space.low
    max_ = self._env.observation_space.high
    observ = 2 * (observ - min_) / (max_ - min_) - 1
    return observ

  def _is_finite(self, space):
    return np.isfinite(space.low).all() and np.isfinite(space.high).all()


class ClipAction(object):
  """Clip out of range actions to the action space of the environment."""

  def __init__(self, env):
    self._env = env

  def __getattr__(self, name):
    return getattr(self._env, name)

  @property
  def action_space(self):
    shape = self._env.action_space.shape
    return gym.spaces.Box(-np.inf * np.ones(shape), np.inf * np.ones(shape))

  def step(self, action):
    action_space = self._env.action_space
    action = np.clip(action, action_space.low, action_space.high)
    return self._env.step(action)


class LimitDuration(object):
  """End episodes after specified number of steps."""

  def __init__(self, env, duration):
    self._env = env
    self._duration = duration
    self._step = None

  def __getattr__(self, name):
    return getattr(self._env, name)

  def step(self, action):
    if self._step is None:
      raise RuntimeError('Must reset environment.')
    observ, reward, done, info = self._env.step(action)
    self._step += 1
    if self._step >= self._duration:
      done = True
      self._step = None
    return observ, reward, done, info

  def reset(self):
    self._step = 0
    return self._env.reset()


class ExternalProcess(object):
  """Step environment in a separate process for lock free paralellism."""

  # Message types for communication via the pipe.
  _ACCESS = 1
  _CALL = 2
  _RESULT = 3
  _EXCEPTION = 4
  _CLOSE = 5

  def __init__(self, constructor):
    """Step environment in a separate process for lock free paralellism.

    The environment will be created in the external process by calling the
    specified callable. This can be an environment class, or a function
    creating the environment and potentially wrapping it. The returned
    environment should not access global variables.

    Args:
      constructor: Callable that creates and returns an OpenAI gym environment.

    Attributes:
      observation_space: The cached observation space of the environment.
      action_space: The cached action space of the environment.
    """
    self._conn, conn = multiprocessing.Pipe()
    self._process = multiprocessing.Process(target=self._worker, args=(constructor, conn))
    atexit.register(self.close)
    self._process.start()
    self._observ_space = None
    self._action_space = None

  @property
  def observation_space(self):
    if not self._observ_space:
      self._observ_space = self.__getattr__('observation_space')
    return self._observ_space

  @property
  def action_space(self):
    if not self._action_space:
      self._action_space = self.__getattr__('action_space')
    return self._action_space

  def __getattr__(self, name):
    """Request an attribute from the environment.

    Note that this involves communication with the external process, so it can
    be slow.

    Args:
      name: Attribute to access.

    Returns:
      Value of the attribute.
    """
    self._conn.send((self._ACCESS, name))
    return self._receive()

  def call(self, name, *args, **kwargs):
    """Asynchronously call a method of the external environment.

    Args:
      name: Name of the method to call.
      *args: Positional arguments to forward to the method.
      **kwargs: Keyword arguments to forward to the method.

    Returns:
      Promise object that blocks and provides the return value when called.
    """
    payload = name, args, kwargs
    self._conn.send((self._CALL, payload))
    return self._receive

  def close(self):
    """Send a close message to the external process and join it."""
    try:
      self._conn.send((self._CLOSE, None))
      self._conn.close()
    except IOError:
      # The connection was already closed.
      pass
    self._process.join()

  def step(self, action, blocking=True):
    """Step the environment.

    Args:
      action: The action to apply to the environment.
      blocking: Whether to wait for the result.

    Returns:
      Transition tuple when blocking, otherwise callable that returns the
      transition tuple.
    """
    promise = self.call('step', action)
    if blocking:
      return promise()
    else:
      return promise

  def reset(self, blocking=True):
    """Reset the environment.

    Args:
      blocking: Whether to wait for the result.

    Returns:
      New observation when blocking, otherwise callable that returns the new
      observation.
    """
    promise = self.call('reset')
    if blocking:
      return promise()
    else:
      return promise

  def _receive(self):
    """Wait for a message from the worker process and return its payload.

    Raises:
      Exception: An exception was raised inside the worker process.
      KeyError: The reveived message is of an unknown type.

    Returns:
      Payload object of the message.
    """
    message, payload = self._conn.recv()
    # Re-raise exceptions in the main process.
    if message == self._EXCEPTION:
      stacktrace = payload
      raise Exception(stacktrace)
    if message == self._RESULT:
      return payload
    raise KeyError('Received message of unexpected type {}'.format(message))

  def _worker(self, constructor, conn):
    """The process waits for actions and sends back environment results.

    Args:
      constructor: Constructor for the OpenAI Gym environment.
      conn: Connection for communication to the main process.
    """
    try:
      env = constructor()
      while True:
        try:
          # Only block for short times to have keyboard exceptions be raised.
          if not conn.poll(0.1):
            continue
          message, payload = conn.recv()
        except (EOFError, KeyboardInterrupt):
          break
        if message == self._ACCESS:
          name = payload
          result = getattr(env, name)
          conn.send((self._RESULT, result))
          continue
        if message == self._CALL:
          name, args, kwargs = payload
          result = getattr(env, name)(*args, **kwargs)
          conn.send((self._RESULT, result))
          continue
        if message == self._CLOSE:
          assert payload is None
          break
        raise KeyError('Received message of unknown type {}'.format(message))
    except Exception:  # pylint: disable=broad-except
      stacktrace = ''.join(traceback.format_exception(*sys.exc_info()))
      tf.logging.error('Error in environment process: {}'.format(stacktrace))
      conn.send((self._EXCEPTION, stacktrace))
    conn.close()


class ConvertTo32Bit(object):
  """Convert data types of an OpenAI Gym environment to 32 bit."""

  def __init__(self, env):
    """Convert data types of an OpenAI Gym environment to 32 bit.

    Args:
      env: OpenAI Gym environment.
    """
    self._env = env

  def __getattr__(self, name):
    """Forward unimplemented attributes to the original environment.

    Args:
      name: Attribute that was accessed.

    Returns:
      Value behind the attribute name in the wrapped environment.
    """
    return getattr(self._env, name)

  def step(self, action):
    """Forward action to the wrapped environment.

    Args:
      action: Action to apply to the environment.

    Raises:
      ValueError: Invalid action.

    Returns:
      Converted observation, converted reward, done flag, and info object.
    """
    observ, reward, done, info = self._env.step(action)
    observ = self._convert_observ(observ)
    reward = self._convert_reward(reward)
    return observ, reward, done, info

  def reset(self):
    """Reset the environment and convert the resulting observation.

    Returns:
      Converted observation.
    """
    observ = self._env.reset()
    observ = self._convert_observ(observ)
    return observ

  def _convert_observ(self, observ):
    """Convert the observation to 32 bits.

    Args:
      observ: Numpy observation.

    Raises:
      ValueError: Observation contains infinite values.

    Returns:
      Numpy observation with 32-bit data type.
    """
    if not np.isfinite(observ).all():
      raise ValueError('Infinite observation encountered.')
    if observ.dtype == np.float64:
      return observ.astype(np.float32)
    if observ.dtype == np.int64:
      return observ.astype(np.int32)
    return observ

  def _convert_reward(self, reward):
    """Convert the reward to 32 bits.

    Args:
      reward: Numpy reward.

    Raises:
      ValueError: Rewards contain infinite values.

    Returns:
      Numpy reward with 32-bit data type.
    """
    if not np.isfinite(reward).all():
      raise ValueError('Infinite reward encountered.')
    return np.array(reward, dtype=np.float32)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vector.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='vector.proto',
  package='robotics.messages',
  syntax='proto3',
  serialized_options=b'\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0cvector.proto\x12\x11robotics.messages\"6\n\x08Vector4d\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\x12\t\n\x01w\x18\x04 \x01(\x01\"6\n\x08Vector4f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\x12\t\n\x01w\x18\x04 \x01(\x02\"6\n\x08Vector4i\x12\t\n\x01x\x18\x01 \x01(\x03\x12\t\n\x01y\x18\x02 \x01(\x03\x12\t\n\x01z\x18\x03 \x01(\x03\x12\t\n\x01w\x18\x04 \x01(\x03\"+\n\x08Vector3d\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\"+\n\x08Vector3f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"+\n\x08Vector3i\x12\t\n\x01x\x18\x01 \x01(\x03\x12\t\n\x01y\x18\x02 \x01(\x03\x12\t\n\x01z\x18\x03 \x01(\x03\" \n\x08Vector2d\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\" \n\x08Vector2f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\" \n\x08Vector2i\x12\t\n\x01x\x18\x01 \x01(\x03\x12\t\n\x01y\x18\x02 \x01(\x03\"\x1b\n\x07Vectord\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x01\x42\x02\x10\x01\"\x1b\n\x07Vectorf\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x02\x42\x02\x10\x01\"\x1b\n\x07Vectori\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x03\x42\x02\x10\x01\x42\x03\xf8\x01\x01\x62\x06proto3'
)




_VECTOR4D = _descriptor.Descriptor(
  name='Vector4d',
  full_name='robotics.messages.Vector4d',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector4d.x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector4d.y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.Vector4d.z', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='w', full_name='robotics.messages.Vector4d.w', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=89,
)


_VECTOR4F = _descriptor.Descriptor(
  name='Vector4f',
  full_name='robotics.messages.Vector4f',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector4f.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector4f.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.Vector4f.z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='w', full_name='robotics.messages.Vector4f.w', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=145,
)


_VECTOR4I = _descriptor.Descriptor(
  name='Vector4i',
  full_name='robotics.messages.Vector4i',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector4i.x', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector4i.y', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.Vector4i.z', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='w', full_name='robotics.messages.Vector4i.w', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=147,
  serialized_end=201,
)


_VECTOR3D = _descriptor.Descriptor(
  name='Vector3d',
  full_name='robotics.messages.Vector3d',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector3d.x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector3d.y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.Vector3d.z', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=203,
  serialized_end=246,
)


_VECTOR3F = _descriptor.Descriptor(
  name='Vector3f',
  full_name='robotics.messages.Vector3f',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector3f.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector3f.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.Vector3f.z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=248,
  serialized_end=291,
)


_VECTOR3I = _descriptor.Descriptor(
  name='Vector3i',
  full_name='robotics.messages.Vector3i',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector3i.x', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector3i.y', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.Vector3i.z', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=293,
  serialized_end=336,
)


_VECTOR2D = _descriptor.Descriptor(
  name='Vector2d',
  full_name='robotics.messages.Vector2d',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector2d.x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector2d.y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=338,
  serialized_end=370,
)


_VECTOR2F = _descriptor.Descriptor(
  name='Vector2f',
  full_name='robotics.messages.Vector2f',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector2f.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector2f.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=372,
  serialized_end=404,
)


_VECTOR2I = _descriptor.Descriptor(
  name='Vector2i',
  full_name='robotics.messages.Vector2i',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.Vector2i.x', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.Vector2i.y', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=406,
  serialized_end=438,
)


_VECTORD = _descriptor.Descriptor(
  name='Vectord',
  full_name='robotics.messages.Vectord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='robotics.messages.Vectord.data', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=440,
  serialized_end=467,
)


_VECTORF = _descriptor.Descriptor(
  name='Vectorf',
  full_name='robotics.messages.Vectorf',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='robotics.messages.Vectorf.data', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=469,
  serialized_end=496,
)


_VECTORI = _descriptor.Descriptor(
  name='Vectori',
  full_name='robotics.messages.Vectori',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='robotics.messages.Vectori.data', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=498,
  serialized_end=525,
)

DESCRIPTOR.message_types_by_name['Vector4d'] = _VECTOR4D
DESCRIPTOR.message_types_by_name['Vector4f'] = _VECTOR4F
DESCRIPTOR.message_types_by_name['Vector4i'] = _VECTOR4I
DESCRIPTOR.message_types_by_name['Vector3d'] = _VECTOR3D
DESCRIPTOR.message_types_by_name['Vector3f'] = _VECTOR3F
DESCRIPTOR.message_types_by_name['Vector3i'] = _VECTOR3I
DESCRIPTOR.message_types_by_name['Vector2d'] = _VECTOR2D
DESCRIPTOR.message_types_by_name['Vector2f'] = _VECTOR2F
DESCRIPTOR.message_types_by_name['Vector2i'] = _VECTOR2I
DESCRIPTOR.message_types_by_name['Vectord'] = _VECTORD
DESCRIPTOR.message_types_by_name['Vectorf'] = _VECTORF
DESCRIPTOR.message_types_by_name['Vectori'] = _VECTORI
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Vector4d = _reflection.GeneratedProtocolMessageType('Vector4d', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR4D,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector4d)
  })
_sym_db.RegisterMessage(Vector4d)

Vector4f = _reflection.GeneratedProtocolMessageType('Vector4f', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR4F,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector4f)
  })
_sym_db.RegisterMessage(Vector4f)

Vector4i = _reflection.GeneratedProtocolMessageType('Vector4i', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR4I,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector4i)
  })
_sym_db.RegisterMessage(Vector4i)

Vector3d = _reflection.GeneratedProtocolMessageType('Vector3d', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR3D,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector3d)
  })
_sym_db.RegisterMessage(Vector3d)

Vector3f = _reflection.GeneratedProtocolMessageType('Vector3f', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR3F,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector3f)
  })
_sym_db.RegisterMessage(Vector3f)

Vector3i = _reflection.GeneratedProtocolMessageType('Vector3i', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR3I,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector3i)
  })
_sym_db.RegisterMessage(Vector3i)

Vector2d = _reflection.GeneratedProtocolMessageType('Vector2d', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR2D,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector2d)
  })
_sym_db.RegisterMessage(Vector2d)

Vector2f = _reflection.GeneratedProtocolMessageType('Vector2f', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR2F,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector2f)
  })
_sym_db.RegisterMessage(Vector2f)

Vector2i = _reflection.GeneratedProtocolMessageType('Vector2i', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR2I,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vector2i)
  })
_sym_db.RegisterMessage(Vector2i)

Vectord = _reflection.GeneratedProtocolMessageType('Vectord', (_message.Message,), {
  'DESCRIPTOR' : _VECTORD,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vectord)
  })
_sym_db.RegisterMessage(Vectord)

Vectorf = _reflection.GeneratedProtocolMessageType('Vectorf', (_message.Message,), {
  'DESCRIPTOR' : _VECTORF,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vectorf)
  })
_sym_db.RegisterMessage(Vectorf)

Vectori = _reflection.GeneratedProtocolMessageType('Vectori', (_message.Message,), {
  'DESCRIPTOR' : _VECTORI,
  '__module__' : 'vector_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Vectori)
  })
_sym_db.RegisterMessage(Vectori)


DESCRIPTOR._options = None
_VECTORD.fields_by_name['data']._options = None
_VECTORF.fields_by_name['data']._options = None
_VECTORI.fields_by_name['data']._options = None
# @@protoc_insertion_point(module_scope)



# Lint as: python3
"""Defines the minitaur robot related constants and URDF specs."""

import collections
import math

import gin

MINITAUR_URDF_PATH = "quadruped/minitaur_rainbow_dash.urdf"

INIT_POSITION = (0, 0, 0.2)
INIT_RACK_POSITION = (0, 0, 1)
INIT_ORIENTATION_QUAT = (0, 0, 0, 1)
INIT_ORIENTATION_RPY = (0, 0, 0)

NUM_LEGS = 4

JOINT_NAMES = ("motor_front_leftL_joint", "motor_front_leftR_joint",
               "motor_back_leftL_joint", "motor_back_leftR_joint",
               "motor_front_rightL_joint", "motor_front_rightR_joint",
               "motor_back_rightL_joint", "motor_back_rightR_joint")

INIT_JOINT_ANGLES = collections.OrderedDict(
    zip(JOINT_NAMES, [math.pi / 2, math.pi / 2] * NUM_LEGS))

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_DIRECTIONS = collections.OrderedDict(
    zip(JOINT_NAMES, (-1, -1, -1, -1, 1, 1, 1, 1)))

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_OFFSETS = collections.OrderedDict(
    zip(JOINT_NAMES, (0, 0, 0, 0, 0, 0, 0, 0)))

LEG_ORDER = ["front_left", "back_left", "front_right", "back_right"]

END_EFFECTOR_NAMES = (
    "knee_front_rightR_joint",
    "knee_front_leftL_joint",
    "knee_back_rightR_joint",
    "knee_back_leftL_joint",
)

MOTOR_NAMES = JOINT_NAMES
MOTOR_GROUP = collections.OrderedDict((("body_motors", JOINT_NAMES),))

KNEE_CONSTRAINT_POINT_LONG = [0, 0.0045, 0.088]
KNEE_CONSTRAINT_POINT_SHORT = [0, 0.0045, 0.100]

# Add the gin constants to be used for gin binding in config.
gin.constant("minitaur_constants.MINITAUR_URDF_PATH", MINITAUR_URDF_PATH)
gin.constant("minitaur_constants.MINITAUR_INIT_POSITION", INIT_POSITION)
gin.constant("minitaur_constants.MINITAUR_INIT_ORIENTATION_QUAT",
             INIT_ORIENTATION_QUAT)
gin.constant("minitaur_constants.MINITAUR_INIT_ORIENTATION_RPY",
             INIT_ORIENTATION_RPY)
gin.constant("minitaur_constants.MINITAUR_INIT_JOINT_ANGLES", INIT_JOINT_ANGLES)
gin.constant("minitaur_constants.MINITAUR_JOINT_DIRECTIONS", JOINT_DIRECTIONS)
gin.constant("minitaur_constants.MINITAUR_JOINT_OFFSETS", JOINT_OFFSETS)
gin.constant("minitaur_constants.MINITAUR_MOTOR_NAMES", MOTOR_NAMES)
gin.constant("minitaur_constants.MINITAUR_END_EFFECTOR_NAMES",
             END_EFFECTOR_NAMES)
gin.constant("minitaur_constants.MINITAUR_MOTOR_GROUP", MOTOR_GROUP)



"""A wrapped MinitaurGymEnv with a built-in controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import gin


@gin.configurable
class TrajectoryGeneratorWrapperEnv(object):
  """A wrapped LocomotionGymEnv with a built-in trajectory generator."""

  def __init__(self, gym_env, trajectory_generator):
    """Initialzes the wrapped env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
      trajectory_generator: A trajectory_generator that can potentially modify
        the action and observation. Typticall generators includes the PMTG and
        openloop signals. Expected to have get_action and get_observation
        interfaces.

    Raises:
      ValueError if the controller does not implement get_action and
      get_observation.

    """
    self._gym_env = gym_env
    if not hasattr(trajectory_generator, 'get_action') or not hasattr(
        trajectory_generator, 'get_observation'):
      raise ValueError(
          'The controller does not have the necessary interface(s) implemented.'
      )

    self._trajectory_generator = trajectory_generator

    # The trajectory generator can subsume the action/observation space.
    if hasattr(trajectory_generator, 'observation_space'):
      self.observation_space = self._trajectory_generator.observation_space

    if hasattr(trajectory_generator, 'action_space'):
      self.action_space = self._trajectory_generator.action_space

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def _modify_observation(self, observation):
    return self._trajectory_generator.get_observation(observation)

  def reset(self, initial_motor_angles=None, reset_duration=1.0):
    if getattr(self._trajectory_generator, 'reset'):
      self._trajectory_generator.reset()
    observation = self._gym_env.reset(initial_motor_angles, reset_duration)
    return self._modify_observation(observation)

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array. The input action from an NN agent.

    Returns:
      The tuple containing the modified observation, the reward, the epsiode end
      indicator.

    Raises:
      ValueError if input action is None.

    """

    if action is None:
      raise ValueError('Action cannot be None')

    new_action = self._trajectory_generator.get_action(
        self._gym_env.robot.GetTimeSinceReset(), action)

    original_observation, reward, done, _ = self._gym_env.step(new_action)

    return self._modify_observation(original_observation), reward, done, _



from .scene_abstract import SingleRobotEmptyScene
from .env_bases import MJCFBaseBulletEnv
import numpy as np
from robot_manipulators import Reacher, Pusher, Striker, Thrower


class ReacherBulletEnv(MJCFBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Reacher()
    MJCFBaseBulletEnv.__init__(self, self.robot, render)

  def create_single_player_scene(self, bullet_client):
    return SingleRobotEmptyScene(bullet_client, gravity=0.0, timestep=0.0165, frame_skip=1)

  def step(self, a):
    assert (not self.scene.multiplayer)
    self.robot.apply_action(a)
    self.scene.global_step()

    state = self.robot.calc_state()  # sets self.to_target_vec

    potential_old = self.potential
    self.potential = self.robot.calc_potential()

    electricity_cost = (
        -0.10 * (np.abs(a[0] * self.robot.theta_dot) + np.abs(a[1] * self.robot.gamma_dot)
                )  # work torque*angular_velocity
        - 0.01 * (np.abs(a[0]) + np.abs(a[1]))  # stall torque require some energy
    )
    stuck_joint_cost = -0.1 if np.abs(np.abs(self.robot.gamma) - 1) < 0.01 else 0.0
    self.rewards = [
        float(self.potential - potential_old),
        float(electricity_cost),
        float(stuck_joint_cost)
    ]
    self.HUD(state, a, False)
    return state, sum(self.rewards), False, {}

  def camera_adjust(self):
    x, y, z = self.robot.fingertip.pose().xyz()
    x *= 0.5
    y *= 0.5
    self.camera.move_and_look_at(0.3, 0.3, 0.3, x, y, z)


class PusherBulletEnv(MJCFBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Pusher()
    MJCFBaseBulletEnv.__init__(self, self.robot, render)

  def create_single_player_scene(self, bullet_client):
    return SingleRobotEmptyScene(bullet_client, gravity=9.81, timestep=0.0020, frame_skip=5)

  def step(self, a):
    self.robot.apply_action(a)
    self.scene.global_step()

    state = self.robot.calc_state()  # sets self.to_target_vec

    potential_old = self.potential
    self.potential = self.robot.calc_potential()

    joint_vel = np.array([
        self.robot.shoulder_pan_joint.get_velocity(),
        self.robot.shoulder_lift_joint.get_velocity(),
        self.robot.upper_arm_roll_joint.get_velocity(),
        self.robot.elbow_flex_joint.get_velocity(),
        self.robot.upper_arm_roll_joint.get_velocity(),
        self.robot.wrist_flex_joint.get_velocity(),
        self.robot.wrist_roll_joint.get_velocity()
    ])

    action_product = np.matmul(np.abs(a), np.abs(joint_vel))
    action_sum = np.sum(a)

    electricity_cost = (
        -0.10 * action_product  # work torque*angular_velocity
        - 0.01 * action_sum  # stall torque require some energy
    )

    stuck_joint_cost = 0
    for j in self.robot.ordered_joints:
      if np.abs(j.current_relative_position()[0]) - 1 < 0.01:
        stuck_joint_cost += -0.1

    self.rewards = [
        float(self.potential - potential_old),
        float(electricity_cost),
        float(stuck_joint_cost)
    ]
    self.HUD(state, a, False)
    return state, sum(self.rewards), False, {}

  def calc_potential(self):
    return -100 * np.linalg.norm(self.to_target_vec)

  def camera_adjust(self):
    x, y, z = self.robot.fingertip.pose().xyz()
    x *= 0.5
    y *= 0.5
    self.camera.move_and_look_at(0.3, 0.3, 0.3, x, y, z)


class StrikerBulletEnv(MJCFBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Striker()
    MJCFBaseBulletEnv.__init__(self, self.robot, render)
    self._striked = False
    self._min_strike_dist = np.inf
    self.strike_threshold = 0.1

  def create_single_player_scene(self, bullet_client):
    return SingleRobotEmptyScene(bullet_client, gravity=9.81, timestep=0.0020, frame_skip=5)

  def step(self, a):
    self.robot.apply_action(a)
    self.scene.global_step()

    state = self.robot.calc_state()  # sets self.to_target_vec

    potential_old = self.potential
    self.potential = self.robot.calc_potential()

    joint_vel = np.array([
        self.robot.shoulder_pan_joint.get_velocity(),
        self.robot.shoulder_lift_joint.get_velocity(),
        self.robot.upper_arm_roll_joint.get_velocity(),
        self.robot.elbow_flex_joint.get_velocity(),
        self.robot.upper_arm_roll_joint.get_velocity(),
        self.robot.wrist_flex_joint.get_velocity(),
        self.robot.wrist_roll_joint.get_velocity()
    ])

    action_product = np.matmul(np.abs(a), np.abs(joint_vel))
    action_sum = np.sum(a)

    electricity_cost = (
        -0.10 * action_product  # work torque*angular_velocity
        - 0.01 * action_sum  # stall torque require some energy
    )

    stuck_joint_cost = 0
    for j in self.robot.ordered_joints:
      if np.abs(j.current_relative_position()[0]) - 1 < 0.01:
        stuck_joint_cost += -0.1

    dist_object_finger = self.robot.object.pose().xyz() - self.robot.fingertip.pose().xyz()
    reward_dist_vec = self.robot.object.pose().xyz() - self.robot.target.pose().xyz(
    )  # TODO: Should the object and target really belong to the robot? Maybe split this off

    self._min_strike_dist = min(self._min_strike_dist, np.linalg.norm(reward_dist_vec))

    if np.linalg.norm(dist_object_finger) < self.strike_threshold:
      self._striked = True
      self._strike_pos = self.robot.fingertip.pose().xyz()

    if self._striked:
      reward_near_vec = self.robot.object.pose().xyz() - self._strike_pos
    else:
      reward_near_vec = self.robot.object.pose().xyz() - self.robot.fingertip.pose().xyz()

    reward_near = -np.linalg.norm(reward_near_vec)

    reward_dist = -np.linalg.norm(self._min_strike_dist)
    reward_ctrl = -np.square(a).sum()
    self.rewards = [
        float(self.potential - potential_old),
        float(electricity_cost),
        float(stuck_joint_cost), 3 * reward_dist, 0.1 * reward_ctrl, 0.5 * reward_near
    ]
    self.HUD(state, a, False)
    return state, sum(self.rewards), False, {}

  def calc_potential(self):
    return -100 * np.linalg.norm(self.to_target_vec)

  def camera_adjust(self):
    x, y, z = self.robot.fingertip.pose().xyz()
    x *= 0.5
    y *= 0.5
    self.camera.move_and_look_at(0.3, 0.3, 0.3, x, y, z)


class ThrowerBulletEnv(MJCFBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Thrower()
    MJCFBaseBulletEnv.__init__(self, self.robot, render)

  def create_single_player_scene(self, bullet_client):
    return SingleRobotEmptyScene(bullet_client, gravity=0.0, timestep=0.0020, frame_skip=5)

  def step(self, a):
    self.robot.apply_action(a)
    self.scene.global_step()
    state = self.robot.calc_state()  # sets self.to_target_vec

    potential_old = self.potential
    self.potential = self.robot.calc_potential()

    joint_vel = np.array([
        self.robot.shoulder_pan_joint.get_velocity(),
        self.robot.shoulder_lift_joint.get_velocity(),
        self.robot.upper_arm_roll_joint.get_velocity(),
        self.robot.elbow_flex_joint.get_velocity(),
        self.robot.upper_arm_roll_joint.get_velocity(),
        self.robot.wrist_flex_joint.get_velocity(),
        self.robot.wrist_roll_joint.get_velocity()
    ])

    action_product = np.matmul(np.abs(a), np.abs(joint_vel))
    action_sum = np.sum(a)

    electricity_cost = (
        -0.10 * action_product  # work torque*angular_velocity
        - 0.01 * action_sum  # stall torque require some energy
    )

    stuck_joint_cost = 0
    for j in self.robot.ordered_joints:
      if np.abs(j.current_relative_position()[0]) - 1 < 0.01:
        stuck_joint_cost += -0.1

    object_xy = self.robot.object.pose().xyz()[:2]
    target_xy = self.robot.target.pose().xyz()[:2]

    if not self.robot._object_hit_ground and self.robot.object.pose().xyz(
    )[2] < -0.25:  # TODO: Should the object and target really belong to the robot? Maybe split this off
      self.robot._object_hit_ground = True
      self.robot._object_hit_location = self.robot.object.pose().xyz()

    if self.robot._object_hit_ground:
      object_hit_xy = self.robot._object_hit_location[:2]
      reward_dist = -np.linalg.norm(object_hit_xy - target_xy)
    else:
      reward_dist = -np.linalg.norm(object_xy - target_xy)
    reward_ctrl = -np.square(a).sum()

    self.rewards = [
        float(self.potential - potential_old),
        float(electricity_cost),
        float(stuck_joint_cost), reward_dist, 0.002 * reward_ctrl
    ]
    self.HUD(state, a, False)
    return state, sum(self.rewards), False, {}

  def camera_adjust(self):
    x, y, z = self.robot.fingertip.pose().xyz()
    x *= 0.5
    y *= 0.5
    self.camera.move_and_look_at(0.3, 0.3, 0.3, x, y, z)



"""Exception types for safety related error."""


class SafetyError(Exception):
  """The base safety exception."""
  pass


class OutOfBoundError(SafetyError):
  """Rasied when values like motor position or velocity is out of bound."""
  pass



# Lint as: python3
"""Classic model predictive control methods."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
from typing import Sequence
import numpy as np

_MAX_ABS_RPY = 0.3
_MAX_ABS_ANGULAR_VELOCITY = math.pi

# The center of mass torque is computed using a simple PD control: tau = -KP *
# delta_rotation - KD * delta_angular_velocity
_TORQUE_KP = 2000
_TORQUE_KD = 150

# For center of mass force, we only need to track position in the z direction
# (i.e. maintain the body height), and speed in x-y plane.
_FORCE_KP = 500
_FORCE_KD = 200


def compute_contact_force_projection_matrix(
    foot_positions_in_com_frame: Sequence[Sequence[float]],
    stance_foot_ids: Sequence[int],
) -> np.ndarray:
  r"""Computes the 6 x 3n matrix to map contact force to com dynamics.

  This is essentially the vectorized rhs of com dynamics equation:
      ma = \sum f
      I\omega_dot = \sum r \cross f

  where the summation if over all feet in contact with ground.

  Caveats: Current we have not taken the com rotation into account as we assume
  the com rotation would be small. Ideally we should rotate the foot_positions
  to a world frame centered at com. Also, since absolute yaw angles are not
  accurate dute to drifting, we should use (roll, pitch, 0) to do the foot
  position projection. This feature will be quite useful for MPC.
  TODO(b/143378213): Fix this.

  Args:
    foot_positions_in_com_frame: the local position of each foot.
    stance_foot_ids: The stance foot to be used to assemble the matrix.

  Returns:
    The contact force projection matrix.

  """
  jacobians = []
  for foot_id in stance_foot_ids:
    jv = np.identity(3)
    foot_position = foot_positions_in_com_frame[foot_id]
    x, y, z = foot_position[:3]
    jw = np.array(((0, -z, y), (z, 0, -x), (-y, x, 0)))
    jacobians.append(np.vstack((jv, jw)))

  return np.hstack(jacobians)


def plan_foot_contact_force(
    mass: float,
    inertia: np.ndarray,
    com_position: np.ndarray,
    com_velocity: np.ndarray,
    com_roll_pitch_yaw: np.ndarray,
    com_angular_velocity: np.ndarray,
    foot_positions_in_com_frame: Sequence[Sequence[float]],
    foot_contact_state: Sequence[bool],
    desired_com_position: np.ndarray,
    desired_com_velocity: np.ndarray,
    desired_com_roll_pitch_yaw: np.ndarray,
    desired_com_angular_velocity: np.ndarray,
):
  """Plan the foot contact forces using robot states.

  TODO(b/143382305): Wrap this interface in a MPC class so we can use other
  planning algorithms.

  Args:
    mass: The total mass of the robot.
    inertia: The diagnal elements [Ixx, Iyy, Izz] of the robot.
    com_position: Center of mass position in world frame. Usually we cannot
      accurrately obtain this without motion capture.
    com_velocity: Center of mass velocity in world frame.
    com_roll_pitch_yaw: Center of mass rotation wrt world frame in euler angles.
    com_angular_velocity: The angular velocity (roll_dot, pitch_dot, yaw_dot).
    foot_positions_in_com_frame: The position of all feet/toe joints in the body
      frame.
    foot_contact_state: Indicates if a foot is in contact with the ground.
    desired_com_position: We usually just care about the body height.
    desired_com_velocity: In world frame.
    desired_com_roll_pitch_yaw: We usually care about roll and pitch, since yaw
      measurement can be unreliable.
    desired_com_angular_velocity: Roll and pitch change rate are usually zero.
      Yaw rate is the turning speed of the robot.

  Returns:
    The desired stance foot contact forces.
  """
  del inertia
  del com_position
  body_height = []
  stance_foot_ids = []
  for foot_id, foot_position in enumerate(foot_positions_in_com_frame):
    if not foot_contact_state[foot_id]:
      continue
    stance_foot_ids.append(foot_id)
    body_height.append(foot_position[2])

  avg_bogy_height = abs(sum(body_height) / len(body_height))

  rpy = com_roll_pitch_yaw
  rpy[:2] = np.clip(rpy[:2], -_MAX_ABS_RPY, _MAX_ABS_RPY)
  rpy_dot = com_angular_velocity
  rpy_dot = np.clip(rpy_dot, -_MAX_ABS_ANGULAR_VELOCITY,
                    _MAX_ABS_ANGULAR_VELOCITY)

  com_torque = -avg_bogy_height * (
      _TORQUE_KP * (rpy - desired_com_roll_pitch_yaw) + _TORQUE_KD * rpy_dot)

  # We don't care about the absolute yaw angle in the low level controller.
  # Instead, we stabialize the angular velocity in the z direction.
  com_torque[2] = -avg_bogy_height * _TORQUE_KD * (
      rpy_dot[2] - desired_com_angular_velocity[2])

  # Track a desired com velocity.
  com_force = -_FORCE_KD * (com_velocity - desired_com_velocity)

  # In the z-direction we also want to track the body height.
  com_force[2] += mass * 9.8 - _FORCE_KP * (
      avg_bogy_height - desired_com_position[2])

  com_force_torque = np.concatenate((com_force, com_torque)).transpose()

  # Map the com force torque to foot contact forces.
  foot_force_to_com = compute_contact_force_projection_matrix(
      foot_positions_in_com_frame, stance_foot_ids)
  all_contact_force = -np.matmul(
      np.linalg.pinv(foot_force_to_com), com_force_torque).transpose()
  contact_force = {}

  for i, foot_id in enumerate(stance_foot_ids):
    contact_force[foot_id] = all_contact_force[3 * i:3 * i + 3]

  return contact_force



"""Simple openloop trajectory generators."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import attr
from gym import spaces
import numpy as np

import gin
from pybullet_envs.minitaur.envs_v2.utilities import laikago_pose_utils
from pybullet_envs.minitaur.envs_v2.utilities import minitaur_pose_utils


@gin.configurable
class MinitaurPoseOffsetGenerator(object):
  """A trajectory generator that return a constant leg pose."""

  def __init__(self,
               init_swing=0,
               init_extension=2.0,
               init_pose=None,
               action_scale=1.0,
               action_limit=0.5):
    """Initializes the controller.

    Args:
      init_swing: the swing of the default pose offset
      init_extension: the extension of the default pose offset
      init_pose: the default pose offset, which is None by default. If not None,
        it will define the default pose offset while ignoring init_swing and
        init_extension.
      action_scale: changes the magnitudes of actions
      action_limit: clips actions
    """
    if init_pose is None:
      self._pose = np.array(
          attr.astuple(
              minitaur_pose_utils.MinitaurPose(
                  swing_angle_0=init_swing,
                  swing_angle_1=init_swing,
                  swing_angle_2=init_swing,
                  swing_angle_3=init_swing,
                  extension_angle_0=init_extension,
                  extension_angle_1=init_extension,
                  extension_angle_2=init_extension,
                  extension_angle_3=init_extension)))
    else:  # Ignore init_swing and init_extension
      self._pose = np.array(init_pose)
    action_high = np.array([action_limit] * minitaur_pose_utils.NUM_MOTORS)
    self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)
    self._action_scale = action_scale

  def reset(self):
    pass

  def get_action(self, current_time=None, input_action=None):
    """Computes the trajectory according to input time and action.

    Args:
      current_time: The time in gym env since reset.
      input_action: A numpy array. The input leg pose from a NN controller.

    Returns:
      A numpy array. The desired motor angles.
    """
    del current_time
    return minitaur_pose_utils.leg_pose_to_motor_angles(
        self._pose + self._action_scale * np.array(input_action))

  def get_observation(self, input_observation):
    """Get the trajectory generator's observation."""

    return input_observation


@gin.configurable
class LaikagoPoseOffsetGenerator(object):
  """A trajectory generator that return constant motor angles."""

  def __init__(
      self,
      init_abduction=laikago_pose_utils.LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
      init_hip=laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
      init_knee=laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE,
      action_limit=0.5,
      ):
    """Initializes the controller."""
    self._pose = np.array(
        attr.astuple(
            laikago_pose_utils.LaikagoPose(
                abduction_angle_0=init_abduction,
                hip_angle_0=init_hip,
                knee_angle_0=init_knee,
                abduction_angle_1=init_abduction,
                hip_angle_1=init_hip,
                knee_angle_1=init_knee,
                abduction_angle_2=init_abduction,
                hip_angle_2=init_hip,
                knee_angle_2=init_knee,
                abduction_angle_3=init_abduction,
                hip_angle_3=init_hip,
                knee_angle_3=init_knee)))
    action_high = np.array([action_limit] * 12)
    self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)

  def reset(self):
    pass

  def get_action(self, current_time=None, input_action=None):
    """Computes the trajectory according to input time and action.

    Args:
      current_time: The time in gym env since reset.
      input_action: A numpy array. The input leg pose from a NN controller.

    Returns:
      A numpy array. The desired motor angles.
    """
    del current_time
    return self._pose + input_action

  def get_observation(self, input_observation):
    """Get the trajectory generator's observation."""

    return input_observation



import pybullet as p
import time

import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.setGravity(0, 0, -10)
p.setPhysicsEngineParameter(enableSAT=1)
p.loadURDF("cube_concave.urdf", [0, 0, -25],
           globalScaling=50,
           useFixedBase=True,
           flags=p.URDF_INITIALIZE_SAT_FEATURES)
p.loadURDF("cube.urdf", [0, 0, 1], globalScaling=1, flags=p.URDF_INITIALIZE_SAT_FEATURES)
p.loadURDF("duck_vhacd.urdf", [1, 0, 1], globalScaling=1, flags=p.URDF_INITIALIZE_SAT_FEATURES)

while (p.isConnected()):
  p.stepSimulation()
  pts = p.getContactPoints()
  #print("num contacts = ", len(pts))
  time.sleep(1. / 240.)



#a mimic joint can act as a gear between two joints
#you can control the gear ratio in magnitude and sign (>0 reverses direction)

import pybullet as p
import time
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.loadURDF("plane.urdf", 0, 0, -2)
wheelA = p.loadURDF("differential/diff_ring.urdf", [0, 0, 0])
for i in range(p.getNumJoints(wheelA)):
  print(p.getJointInfo(wheelA, i))
  p.setJointMotorControl2(wheelA, i, p.VELOCITY_CONTROL, targetVelocity=0, force=0)

c = p.createConstraint(wheelA,
                       1,
                       wheelA,
                       3,
                       jointType=p.JOINT_GEAR,
                       jointAxis=[0, 1, 0],
                       parentFramePosition=[0, 0, 0],
                       childFramePosition=[0, 0, 0])
p.changeConstraint(c, gearRatio=1, maxForce=10000)

c = p.createConstraint(wheelA,
                       2,
                       wheelA,
                       4,
                       jointType=p.JOINT_GEAR,
                       jointAxis=[0, 1, 0],
                       parentFramePosition=[0, 0, 0],
                       childFramePosition=[0, 0, 0])
p.changeConstraint(c, gearRatio=-1, maxForce=10000)

c = p.createConstraint(wheelA,
                       1,
                       wheelA,
                       4,
                       jointType=p.JOINT_GEAR,
                       jointAxis=[0, 1, 0],
                       parentFramePosition=[0, 0, 0],
                       childFramePosition=[0, 0, 0])
p.changeConstraint(c, gearRatio=-1, maxForce=10000)

p.setRealTimeSimulation(1)
while (1):
  p.setGravity(0, 0, -10)
  time.sleep(0.01)
#p.removeConstraint(c)



import pybullet as p
import time
import math
import pybullet_data


def getRayFromTo(mouseX, mouseY):
  width, height, viewMat, projMat, cameraUp, camForward, horizon, vertical, _, _, dist, camTarget = p.getDebugVisualizerCamera(
  )
  camPos = [
      camTarget[0] - dist * camForward[0], camTarget[1] - dist * camForward[1],
      camTarget[2] - dist * camForward[2]
  ]
  farPlane = 10000
  rayForward = [(camTarget[0] - camPos[0]), (camTarget[1] - camPos[1]), (camTarget[2] - camPos[2])]
  invLen = farPlane * 1. / (math.sqrt(rayForward[0] * rayForward[0] + rayForward[1] *
                                      rayForward[1] + rayForward[2] * rayForward[2]))
  rayForward = [invLen * rayForward[0], invLen * rayForward[1], invLen * rayForward[2]]
  rayFrom = camPos
  oneOverWidth = float(1) / float(width)
  oneOverHeight = float(1) / float(height)
  dHor = [horizon[0] * oneOverWidth, horizon[1] * oneOverWidth, horizon[2] * oneOverWidth]
  dVer = [vertical[0] * oneOverHeight, vertical[1] * oneOverHeight, vertical[2] * oneOverHeight]
  rayToCenter = [
      rayFrom[0] + rayForward[0], rayFrom[1] + rayForward[1], rayFrom[2] + rayForward[2]
  ]
  rayTo = [
      rayFrom[0] + rayForward[0] - 0.5 * horizon[0] + 0.5 * vertical[0] + float(mouseX) * dHor[0] -
      float(mouseY) * dVer[0], rayFrom[1] + rayForward[1] - 0.5 * horizon[1] + 0.5 * vertical[1] +
      float(mouseX) * dHor[1] - float(mouseY) * dVer[1], rayFrom[2] + rayForward[2] -
      0.5 * horizon[2] + 0.5 * vertical[2] + float(mouseX) * dHor[2] - float(mouseY) * dVer[2]
  ]
  return rayFrom, rayTo


cid = p.connect(p.SHARED_MEMORY)
if (cid < 0):
  p.connect(p.GUI)

p.setAdditionalSearchPath(pybullet_data.getDataPath())  
p.setPhysicsEngineParameter(numSolverIterations=10)
p.setTimeStep(1. / 120.)
logId = p.startStateLogging(p.STATE_LOGGING_PROFILE_TIMINGS, "visualShapeBench.json")
#useMaximalCoordinates is much faster then the default reduced coordinates (Featherstone)
p.loadURDF("plane100.urdf", useMaximalCoordinates=True)
#disable rendering during creation.
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0)
p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
#disable tinyrenderer, software (CPU) renderer, we don't use it here
p.configureDebugVisualizer(p.COV_ENABLE_TINY_RENDERER, 0)

shift = [0, -0.02, 0]
shift1 = [0, 0.1, 0]
shift2 = [0, 0, 0]

meshScale = [0.1, 0.1, 0.1]
#the visual shape and collision shape can be re-used by all createMultiBody instances (instancing)
visualShapeId = p.createVisualShapeArray(shapeTypes=[p.GEOM_MESH, p.GEOM_BOX],
                                         halfExtents=[[0, 0, 0], [0.1, 0.1, 0.1]],
                                         fileNames=["duck.obj", ""],
                                         visualFramePositions=[
                                             shift1,
                                             shift2,
                                         ],
                                         meshScales=[meshScale, meshScale])
collisionShapeId = p.createCollisionShapeArray(shapeTypes=[p.GEOM_MESH, p.GEOM_BOX],
                                               halfExtents=[[0, 0, 0], [0.1, 0.1, 0.1]],
                                               fileNames=["duck_vhacd.obj", ""],
                                               collisionFramePositions=[
                                                   shift1,
                                                   shift2,
                                               ],
                                               meshScales=[meshScale, meshScale])

rangex = 2
rangey = 2
for i in range(rangex):
  for j in range(rangey):
    mb = p.createMultiBody(baseMass=1,
                           baseInertialFramePosition=[0, 0, 0],
                           baseCollisionShapeIndex=collisionShapeId,
                           baseVisualShapeIndex=visualShapeId,
                           basePosition=[((-rangex / 2) + i * 2) * meshScale[0] * 2,
                                         (-rangey / 2 + j) * meshScale[1] * 4, 1],
                           useMaximalCoordinates=False)
    p.changeVisualShape(mb, -1, rgbaColor=[1, 1, 1, 1])
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)
p.stopStateLogging(logId)
p.setGravity(0, 0, -10)
p.setRealTimeSimulation(1)

colors = [[1, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1], [1, 1, 1, 1]]
currentColor = 0

p.getCameraImage(64, 64, renderer=p.ER_BULLET_HARDWARE_OPENGL)

while (1):

  mouseEvents = p.getMouseEvents()
  for e in mouseEvents:
    if ((e[0] == 2) and (e[3] == 0) and (e[4] & p.KEY_WAS_TRIGGERED)):
      mouseX = e[1]
      mouseY = e[2]
      rayFrom, rayTo = getRayFromTo(mouseX, mouseY)
      rayInfo = p.rayTest(rayFrom, rayTo)
      #p.addUserDebugLine(rayFrom,rayTo,[1,0,0],3)
      for l in range(len(rayInfo)):
        hit = rayInfo[l]
        objectUid = hit[0]
        if (objectUid >= 0):
          #p.removeBody(objectUid)
          p.changeVisualShape(objectUid, -1, rgbaColor=colors[currentColor])
          currentColor += 1
          if (currentColor >= len(colors)):
            currentColor = 0



"""An agent that can restore and run a policy learned by PPO."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf
from pybullet_envs.agents.ppo import normalize
from pybullet_envs.agents import utility


class SimplePPOPolicy(object):
  """A simple PPO policy that is independent to the PPO infrastructure.

  This class restores the policy network from a tensorflow checkpoint that was
  learned from PPO training. The purpose of this class is to conveniently
  visualize a learned policy or deploy the learned policy on real robots without
  need to change the PPO evaluation infrastructure:
  https://cs.corp.google.com/piper///depot/google3/robotics/reinforcement_learning/agents/scripts/visualize.py.
  """

  def __init__(self, sess, env, network, policy_layers, value_layers, checkpoint):
    self.env = env
    self.sess = sess
    observation_size = len(env.observation_space.low)
    action_size = len(env.action_space.low)
    self.observation_placeholder = tf.placeholder(tf.float32, [None, observation_size],
                                                  name="Input")
    self._observ_filter = normalize.StreamingNormalize(self.observation_placeholder[0],
                                                       center=True,
                                                       scale=True,
                                                       clip=5,
                                                       name="normalize_observ")
    self._restore_policy(network,
                         policy_layers=policy_layers,
                         value_layers=value_layers,
                         action_size=action_size,
                         checkpoint=checkpoint)

  def _restore_policy(self, network, policy_layers, value_layers, action_size, checkpoint):
    """Restore the PPO policy from a TensorFlow checkpoint.

    Args:
      network: The neural network definition.
      policy_layers: A tuple specify the number of layers and number of neurons
        of each layer for the policy network.
      value_layers: A tuple specify the number of layers and number of neurons
        of each layer for the value network.
      action_size: The dimension of the action space.
      checkpoint: The checkpoint path.
    """
    observ = self._observ_filter.transform(self.observation_placeholder)
    with tf.variable_scope("network/rnn"):
      self.network = network(policy_layers=policy_layers,
                             value_layers=value_layers,
                             action_size=action_size)

    with tf.variable_scope("temporary"):
      self.last_state = tf.Variable(self.network.zero_state(1, tf.float32), False)
      self.sess.run(self.last_state.initializer)

    with tf.variable_scope("network"):
      (mean_action, _, _), new_state = tf.nn.dynamic_rnn(self.network,
                                                         observ[:, None],
                                                         tf.ones(1),
                                                         self.last_state,
                                                         tf.float32,
                                                         swap_memory=True)
      self.mean_action = mean_action
      self.update_state = self.last_state.assign(new_state)

    saver = utility.define_saver(exclude=(r"temporary/.*",))
    saver.restore(self.sess, checkpoint)

  def get_action(self, observation):
    normalized_observation = self._normalize_observ(observation)
    normalized_action, _ = self.sess.run(
        [self.mean_action, self.update_state],
        feed_dict={self.observation_placeholder: normalized_observation})
    action = self._denormalize_action(normalized_action)
    return action[:, 0]

  def _denormalize_action(self, action):
    min_ = self.env.action_space.low
    max_ = self.env.action_space.high
    action = (action + 1) / 2 * (max_ - min_) + min_
    return action

  def _normalize_observ(self, observ):
    min_ = self.env.observation_space.low
    max_ = self.env.observation_space.high
    observ = 2 * (observ - min_) / (max_ - min_) - 1
    return observ



import pybullet as p
import pybullet_data as pd
import time
p.connect(p.GUI)#, options="--background_color_red=1.0 --background_color_blue=1.0 --background_color_green=1.0")
p.setAdditionalSearchPath(pd.getDataPath())


useFixedBase = True
flags = p.URDF_INITIALIZE_SAT_FEATURES#0#p.URDF_USE_SELF_COLLISION

#plane_pos = [0,0,0]
#plane = p.loadURDF("plane.urdf", plane_pos, flags = flags, useFixedBase=useFixedBase)
table_pos = [0,0,-0.625]
table = p.loadURDF("table/table.urdf", table_pos, flags = flags, useFixedBase=useFixedBase)
xarm = p.loadURDF("xarm/xarm6_robot.urdf", flags = flags, useFixedBase=useFixedBase)

jointIds = []
paramIds = []

for j in range(p.getNumJoints(xarm)):
  p.changeDynamics(xarm, j, linearDamping=0, angularDamping=0)
  info = p.getJointInfo(xarm, j)
  #print(info)
  jointName = info[1]
  jointType = info[2]
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    jointIds.append(j)
    paramIds.append(p.addUserDebugParameter(jointName.decode("utf-8"), -4, 4, 0))
  
skip_cam_frames = 10  

while (1):
	p.stepSimulation()
	for i in range(len(paramIds)):
		c = paramIds[i]
		targetPos = p.readUserDebugParameter(c)
		p.setJointMotorControl2(xarm, jointIds[i], p.POSITION_CONTROL, targetPos, force=5 * 240.)
	skip_cam_frames -= 1
	if (skip_cam_frames<0):
		p.getCameraImage(320,200, renderer=p.ER_BULLET_HARDWARE_OPENGL )
		skip_cam_frames = 10
	time.sleep(1./240.)
	



"""

Code to load a policy and generate rollout data. Adapted from https://github.com/berkeleydeeprlcourse. 
Example usage:
python3 laikago_ars_run_policy.py --expert_policy_file=data/lin_policy_plus_best_10.npz --json_file=data/params.json

"""
import numpy as np
import gym
import time
import pybullet_envs
try:
  import tds_environments
except:
  pass
import json
from arspb.policies import *
import time
import arspb.trained_policies as tp
import os


import os
import gin
from pybullet_envs.minitaur.envs_v2 import env_loader


import pybullet_data as pd



def create_laikago_env(args):
  CONFIG_DIR = pd.getDataPath()+"/configs_v2/"
  CONFIG_FILES = [
      os.path.join(CONFIG_DIR, "base/laikago_with_imu.gin"),
      os.path.join(CONFIG_DIR, "tasks/fwd_task_no_termination.gin"),
      os.path.join(CONFIG_DIR, "wrappers/pmtg_wrapper.gin"),
      os.path.join(CONFIG_DIR, "scenes/simple_scene.gin")
  ]
   
  #gin.bind_parameter("scene_base.SceneBase.data_root", pd.getDataPath()+"/")
  for gin_file in CONFIG_FILES:
    gin.parse_config_file(gin_file)
  gin.bind_parameter("SimulationParameters.enable_rendering", True)
  gin.bind_parameter("terminal_conditions.maxstep_terminal_condition.max_step",
                     10000)
  env = env_loader.load()
  return env

  
def main(argv):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--expert_policy_file', type=str, default="")
    parser.add_argument('--nosleep', action='store_true')

    parser.add_argument('--num_rollouts', type=int, default=20,
                        help='Number of expert rollouts')
    parser.add_argument('--json_file', type=str, default="")
    parser.add_argument('--run_on_robot', action='store_true')
    if len(argv):
      args = parser.parse_args(argv)
    else:
      args = parser.parse_args()

    print('loading and building expert policy')
    if len(args.json_file)==0:
      args.json_file = tp.getDataPath()+"/"+ args.envname+"/params.json"    
    with open(args.json_file) as f:
       params = json.load(f)
    print("params=",params)
    if len(args.expert_policy_file)==0:
      args.expert_policy_file=tp.getDataPath()+"/"+args.envname+"/nn_policy_plus.npz" 
      if not os.path.exists(args.expert_policy_file):
        args.expert_policy_file=tp.getDataPath()+"/"+args.envname+"/lin_policy_plus.npz"
    data = np.load(args.expert_policy_file, allow_pickle=True)

    print('create Laikago gym environment:')
    env = create_laikago_env(args)


    lst = data.files
    weights = data[lst[0]][0]
    mu = data[lst[0]][1]
    print("mu=",mu)
    std = data[lst[0]][2]
    print("std=",std)
        
    ob_dim = env.observation_space.shape[0]
    ac_dim = env.action_space.shape[0]
    ac_lb = env.action_space.low
    ac_ub = env.action_space.high
    
    policy_params={'type': params["policy_type"],
                   'ob_filter':params['filter'],
                   'ob_dim':ob_dim,
                   'ac_dim':ac_dim,
                   'action_lower_bound' : ac_lb,
                   'action_upper_bound' : ac_ub,
    }
    policy_params['weights'] = weights
    policy_params['observation_filter_mean'] = mu
    policy_params['observation_filter_std'] = std
    if params["policy_type"]=="nn":
      print("FullyConnectedNeuralNetworkPolicy")
      policy_sizes_string = params['policy_network_size_list'].split(',')
      print("policy_sizes_string=",policy_sizes_string)
      policy_sizes_list = [int(item) for item in policy_sizes_string]
      print("policy_sizes_list=",policy_sizes_list)
      policy_params['policy_network_size'] = policy_sizes_list
      policy = FullyConnectedNeuralNetworkPolicy(policy_params, update_filter=False)
    else:
      print("LinearPolicy2")
      policy = LinearPolicy2(policy_params, update_filter=False)
    policy.get_weights()
   
    returns = []
    observations = []
    actions = []
    for i in range(args.num_rollouts):
        print('iter', i)
        obs = env.reset()
        done = False
        totalr = 0.
        steps = 0
        while not done:
            action = policy.act(obs)
            observations.append(obs)
            actions.append(action)
            
            #time.sleep(1)
            obs, r, done, _ = env.step(action)
            totalr += r
            steps += 1
            
            #if steps % 100 == 0: print("%i/%i"%(steps, env.spec.timestep_limit))
            #if steps >= env.spec.timestep_limit:
            #    break
        #print("steps=",steps)
        returns.append(totalr)

    print('returns', returns)
    print('mean return', np.mean(returns))
    print('std of return', np.std(returns))
    
if __name__ == '__main__':
    import sys
    main(sys.argv[1:])



"""A wrapped Quadruped with Inverse Kinematics based controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gin
import gym
import numpy as np
from pybullet_envs.minitaur.robots import vision60
from pybullet_envs.minitaur.robots.utilities import kinematics

ACTION_DIM_PER_LEG = 3
ACTION_DIM_BASE = 7
ACTION_DIM_TOTAL = vision60.NUM_LEGS * ACTION_DIM_PER_LEG + ACTION_DIM_BASE


@gin.configurable
class IKBasedWrapperEnv(object):
  """An env using IK to convert toe positions to joint angles."""

  def __init__(self,
               gym_env,
               toe_indices=(3, 7, 11, 15),
               abduction_motor_ids=(0, 3, 6, 9)):
    """Initialzes the wrapped env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
      toe_indices: A list of four pybullet joint indices for the four toes. [3,
        7, 11, 15] for the vision60.
      abduction_motor_ids: A list of four pybullet joint indices for the four
        abuduction motors. [0, 3, 6, 9] for the vision60.
    """
    lower_bound = np.array([-1.0] * ACTION_DIM_TOTAL)
    upper_bound = np.array([1.0] * ACTION_DIM_TOTAL)
    self._gym_env = gym_env
    self.action_space = gym.spaces.Box(lower_bound, upper_bound)
    self._toe_ids = toe_indices
    self._abduction_motor_ids = abduction_motor_ids

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def get_toe_indices(self):
    return self._toe_ids

  def _joint_angles_from_toe_positions_and_base_pose(self, ik_actions):
    """Uses Inverse Kinematics to calculate jont angles.

    Args:
      ik_actions: The action should be local (x, y, z) for each toe. action for
        each leg [x, y, z] in a local frame. This local frame is transformed
        relative to the COM frame using a given translation, and rotation. The
        total action space would be 3 + 4 + 3 * ACTION_DIM_PER_LEG = 16.

    Returns:
      A list of joint angles.
    """
    assert len(ik_actions) == ACTION_DIM_TOTAL

    base_translation_index = vision60.NUM_LEGS * ACTION_DIM_PER_LEG
    base_rotation_index = vision60.NUM_LEGS * ACTION_DIM_PER_LEG + 3

    base_translation = ik_actions[
        base_translation_index:base_translation_index + 3]
    base_rotation = ik_actions[base_rotation_index:base_rotation_index + 4]
    desired_joint_angles = []
    for i in range(vision60.NUM_LEGS):
      local_toe_pos = ik_actions[i * ACTION_DIM_PER_LEG:i * ACTION_DIM_PER_LEG +
                                 ACTION_DIM_PER_LEG]
      leg_joint_ids = [
          self._abduction_motor_ids[i], self._abduction_motor_ids[i] + 1,
          self._abduction_motor_ids[i] + 2
      ]

      desired_joint_angles.extend(
          kinematics.joint_angles_from_link_position(
              robot=self._gym_env.robot,
              link_position=local_toe_pos,
              link_id=self._toe_ids[i],
              joint_ids=leg_joint_ids,
              base_translation=base_translation,
              base_rotation=base_rotation))

    return desired_joint_angles

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array. The input action from an NN agent.

    Returns:
      The tuple containing the modified observation, the reward, the epsiode end
      indicator.

    Raises:
      ValueError if input action is None.

    """
    if action is None:
      raise ValueError('Action cannot be None')

    desired_joint_angles = self._joint_angles_from_toe_positions_and_base_pose(
        ik_actions=action)
    observation, reward, done, _ = self._gym_env.step(desired_joint_angles)

    return observation, reward, done, _



"""This file implements the gym environment of minitaur.

"""
import math
import random

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

from gym import spaces
import numpy as np
from pybullet_envs.minitaur.envs import minitaur_gym_env
import pybullet_data

GOAL_DISTANCE_THRESHOLD = 0.8
GOAL_REWARD = 1000.0
REWARD_SCALING = 1e-3
INIT_BALL_ANGLE = math.pi / 3
INIT_BALL_DISTANCE = 5.0
ACTION_EPS = 0.01


class MinitaurBallGymEnv(minitaur_gym_env.MinitaurGymEnv):
  """The gym environment for the minitaur and a ball.

  It simulates a minitaur (a quadruped robot) and a ball. The state space
  includes the angle and distance of the ball relative to minitaur's base.
  The action space is a steering command. The reward function is based
  on how far the ball is relative to the minitaur's base.

  """

  def __init__(self,
               urdf_root=pybullet_data.getDataPath(),
               self_collision_enabled=True,
               pd_control_enabled=False,
               leg_model_enabled=True,
               on_rack=False,
               render=False):
    """Initialize the minitaur and ball gym environment.

    Args:
      urdf_root: The path to the urdf data folder.
      self_collision_enabled: Whether to enable self collision in the sim.
      pd_control_enabled: Whether to use PD controller for each motor.
      leg_model_enabled: Whether to use a leg motor to reparameterize the action
        space.
      on_rack: Whether to place the minitaur on rack. This is only used to debug
        the walking gait. In this mode, the minitaur's base is hanged midair so
        that its walking gait is clearer to visualize.
      render: Whether to render the simulation.
    """
    super(MinitaurBallGymEnv, self).__init__(urdf_root=urdf_root,
                                             self_collision_enabled=self_collision_enabled,
                                             pd_control_enabled=pd_control_enabled,
                                             leg_model_enabled=leg_model_enabled,
                                             on_rack=on_rack,
                                             render=render)
    self._cam_dist = 2.0
    self._cam_yaw = -70
    self._cam_pitch = -30
    self.action_space = spaces.Box(np.array([-1]), np.array([1]))
    self.observation_space = spaces.Box(np.array([-math.pi, 0]), np.array([math.pi, 100]))

  def reset(self):
    self._ball_id = 0
    super(MinitaurBallGymEnv, self).reset()
    self._init_ball_theta = random.uniform(-INIT_BALL_ANGLE, INIT_BALL_ANGLE)
    self._init_ball_distance = INIT_BALL_DISTANCE
    self._ball_pos = [
        self._init_ball_distance * math.cos(self._init_ball_theta),
        self._init_ball_distance * math.sin(self._init_ball_theta), 1
    ]
    self._ball_id = self._pybullet_client.loadURDF(
        "%s/sphere_with_restitution.urdf" % self._urdf_root, self._ball_pos)
    return self._get_observation()

  def _get_observation(self):
    world_translation_minitaur, world_rotation_minitaur = (
        self._pybullet_client.getBasePositionAndOrientation(self.minitaur.quadruped))
    world_translation_ball, world_rotation_ball = (
        self._pybullet_client.getBasePositionAndOrientation(self._ball_id))
    minitaur_translation_world, minitaur_rotation_world = (self._pybullet_client.invertTransform(
        world_translation_minitaur, world_rotation_minitaur))
    minitaur_translation_ball, _ = (self._pybullet_client.multiplyTransforms(
        minitaur_translation_world, minitaur_rotation_world, world_translation_ball,
        world_rotation_ball))
    distance = math.sqrt(minitaur_translation_ball[0]**2 + minitaur_translation_ball[1]**2)
    angle = math.atan2(minitaur_translation_ball[0], minitaur_translation_ball[1])
    self._observation = [angle - math.pi / 2, distance]
    return self._observation

  def _transform_action_to_motor_command(self, action):
    if self._leg_model_enabled:
      for i, action_component in enumerate(action):
        if not (-self._action_bound - ACTION_EPS <= action_component <=
                self._action_bound + ACTION_EPS):
          raise ValueError("{}th action {} out of bounds.".format(i, action_component))
      action = self._apply_steering_to_locomotion(action)
      action = self.minitaur.ConvertFromLegModel(action)
    return action

  def _apply_steering_to_locomotion(self, action):
    # A hardcoded feedforward walking controller based on sine functions.
    amplitude_swing = 0.5
    amplitude_extension = 0.5
    speed = 200
    steering_amplitude = 0.5 * action[0]
    t = self.minitaur.GetTimeSinceReset()
    a1 = math.sin(t * speed) * (amplitude_swing + steering_amplitude)
    a2 = math.sin(t * speed + math.pi) * (amplitude_swing - steering_amplitude)
    a3 = math.sin(t * speed) * amplitude_extension
    a4 = math.sin(t * speed + math.pi) * amplitude_extension
    action = [a1, a2, a2, a1, a3, a4, a4, a3]
    return action

  def _distance_to_ball(self):
    world_translation_minitaur, _ = (self._pybullet_client.getBasePositionAndOrientation(
        self.minitaur.quadruped))
    world_translation_ball, _ = (self._pybullet_client.getBasePositionAndOrientation(
        self._ball_id))
    distance = math.sqrt((world_translation_ball[0] - world_translation_minitaur[0])**2 +
                         (world_translation_ball[1] - world_translation_minitaur[1])**2)
    return distance

  def _goal_state(self):
    return self._observation[1] < GOAL_DISTANCE_THRESHOLD

  def _reward(self):
    reward = -self._observation[1]
    if self._goal_state():
      reward += GOAL_REWARD
    return reward * REWARD_SCALING

  def _termination(self):
    if self._goal_state():
      return True
    return False



from pybullet_utils import bullet_client
import math


class QuadrupedPoseInterpolator(object):

  def __init__(self):
    pass

  def ComputeLinVel(self, posStart, posEnd, deltaTime):
    vel = [(posEnd[0] - posStart[0]) / deltaTime, (posEnd[1] - posStart[1]) / deltaTime,
           (posEnd[2] - posStart[2]) / deltaTime]
    return vel

  def ComputeAngVel(self, ornStart, ornEnd, deltaTime, bullet_client):
    dorn = bullet_client.getDifferenceQuaternion(ornStart, ornEnd)
    axis, angle = bullet_client.getAxisAngleFromQuaternion(dorn)
    angVel = [(axis[0] * angle) / deltaTime, (axis[1] * angle) / deltaTime,
              (axis[2] * angle) / deltaTime]
    return angVel

  def ComputeAngVelRel(self, ornStart, ornEnd, deltaTime, bullet_client):
    ornStartConjugate = [-ornStart[0], -ornStart[1], -ornStart[2], ornStart[3]]
    pos_diff, q_diff = bullet_client.multiplyTransforms([0, 0, 0], ornStartConjugate, [0, 0, 0],
                                                        ornEnd)
    axis, angle = bullet_client.getAxisAngleFromQuaternion(q_diff)
    angVel = [(axis[0] * angle) / deltaTime, (axis[1] * angle) / deltaTime,
              (axis[2] * angle) / deltaTime]
    return angVel

  def Slerp(self, frameFraction, frameData, frameDataNext, bullet_client):
    keyFrameDuration = frameData[0]
    basePos1Start = [frameData[1], frameData[2], frameData[3]]
    basePos1End = [frameDataNext[1], frameDataNext[2], frameDataNext[3]]
    self._basePos = [
        basePos1Start[0] + frameFraction * (basePos1End[0] - basePos1Start[0]),
        basePos1Start[1] + frameFraction * (basePos1End[1] - basePos1Start[1]),
        basePos1Start[2] + frameFraction * (basePos1End[2] - basePos1Start[2])
    ]
    self._baseLinVel = self.ComputeLinVel(basePos1Start, basePos1End, keyFrameDuration)
    baseOrn1Start = [frameData[5], frameData[6], frameData[7], frameData[4]]
    baseOrn1Next = [frameDataNext[5], frameDataNext[6], frameDataNext[7], frameDataNext[4]]
    self._baseOrn = bullet_client.getQuaternionSlerp(baseOrn1Start, baseOrn1Next, frameFraction)
    self._baseAngVel = self.ComputeAngVel(baseOrn1Start, baseOrn1Next, keyFrameDuration,
                                          bullet_client)

    jointPositions = [
        self._basePos[0], self._basePos[1], self._basePos[2], self._baseOrn[0], self._baseOrn[1],
        self._baseOrn[2], self._baseOrn[3]
    ]
    jointVelocities = [
        self._baseLinVel[0], self._baseLinVel[1], self._baseLinVel[2], self._baseAngVel[0],
        self._baseAngVel[1], self._baseAngVel[2]
    ]

    for j in range(12):
      index = j + 8
      jointPosStart = frameData[index]
      jointPosEnd = frameDataNext[index]
      jointPos = jointPosStart + frameFraction * (jointPosEnd - jointPosStart)
      jointVel = (jointPosEnd - jointPosStart) / keyFrameDuration
      jointPositions.append(jointPos)
      jointVelocities.append(jointVel)
    self._jointPositions = jointPositions
    self._jointVelocities = jointVelocities
    return jointPositions, jointVelocities



import numpy as np
import math
from pybullet_envs.deep_mimic.env.env import Env
from pybullet_envs.deep_mimic.env.action_space import ActionSpace
from pybullet_utils import bullet_client
import time
from pybullet_envs.deep_mimic.env import motion_capture_data
from pybullet_envs.deep_mimic.env import humanoid_stable_pd
import pybullet_data
import pybullet as p1
import random

from enum import Enum

class InitializationStrategy(Enum):
  """Set how the environment is initialized."""
  START = 0
  RANDOM = 1  # random state initialization (RSI)


class PyBulletDeepMimicEnv(Env):

  def __init__(self, arg_parser=None, enable_draw=False, pybullet_client=None,
               time_step=1./240,
               init_strategy=InitializationStrategy.RANDOM):
    super().__init__(arg_parser, enable_draw)
    self._num_agents = 1
    self._pybullet_client = pybullet_client
    self._isInitialized = False
    self._useStablePD = True
    self._arg_parser = arg_parser
    self.timeStep = time_step
    self._init_strategy = init_strategy
    print("Initialization strategy: {:s}".format(init_strategy))
    self.reset()

  def reset(self):

    if not self._isInitialized:
      if self.enable_draw:
        self._pybullet_client = bullet_client.BulletClient(connection_mode=p1.GUI)
        #disable 'GUI' since it slows down a lot on Mac OSX and some other platforms
        self._pybullet_client.configureDebugVisualizer(self._pybullet_client.COV_ENABLE_GUI, 0)
      else:
        self._pybullet_client = bullet_client.BulletClient()

      self._pybullet_client.setAdditionalSearchPath(pybullet_data.getDataPath())
      z2y = self._pybullet_client.getQuaternionFromEuler([-math.pi * 0.5, 0, 0])
      self._planeId = self._pybullet_client.loadURDF("plane_implicit.urdf", [0, 0, 0],
                                                     z2y,
                                                     useMaximalCoordinates=True)
      #print("planeId=",self._planeId)
      self._pybullet_client.configureDebugVisualizer(self._pybullet_client.COV_ENABLE_Y_AXIS_UP, 1)
      self._pybullet_client.setGravity(0, -9.8, 0)

      self._pybullet_client.setPhysicsEngineParameter(numSolverIterations=10)
      self._pybullet_client.changeDynamics(self._planeId, linkIndex=-1, lateralFriction=0.9)

      self._mocapData = motion_capture_data.MotionCaptureData()

      motion_file = self._arg_parser.parse_strings('motion_file')
      print("motion_file=", motion_file[0])

      motionPath = pybullet_data.getDataPath() + "/" + motion_file[0]
      #motionPath = pybullet_data.getDataPath()+"/motions/humanoid3d_backflip.txt"
      self._mocapData.Load(motionPath)
      timeStep = self.timeStep
      useFixedBase = False
      self._humanoid = humanoid_stable_pd.HumanoidStablePD(self._pybullet_client, self._mocapData,
                                                           timeStep, useFixedBase, self._arg_parser)
      self._isInitialized = True

      self._pybullet_client.setTimeStep(timeStep)
      self._pybullet_client.setPhysicsEngineParameter(numSubSteps=1)

      selfCheck = False
      if (selfCheck):
        curTime = 0
        while self._pybullet_client.isConnected():
          self._humanoid.setSimTime(curTime)
          state = self._humanoid.getState()
          #print("state=",state)
          pose = self._humanoid.computePose(self._humanoid._frameFraction)
          for i in range(10):
            curTime += timeStep
            #taus = self._humanoid.computePDForces(pose)
            #self._humanoid.applyPDForces(taus)
            #self._pybullet_client.stepSimulation()
          time.sleep(timeStep)
    #print("numframes = ", self._humanoid._mocap_data.NumFrames())
    #startTime = random.randint(0,self._humanoid._mocap_data.NumFrames()-2)
    
    if self._init_strategy == InitializationStrategy.RANDOM:
      rnrange = 1000
      rn = random.randint(0, rnrange)
      startTime = float(rn) / rnrange * self._humanoid.getCycleTime()
    elif self._init_strategy == InitializationStrategy.START:
      startTime = 0
    
    self.t = startTime

    self._humanoid.setSimTime(startTime)

    self._humanoid.resetPose()
    #this clears the contact points. Todo: add API to explicitly clear all contact points?
    #self._pybullet_client.stepSimulation()
    self._humanoid.resetPose()
    self.needs_update_time = self.t - 1  #force update

  def get_num_agents(self):
    return self._num_agents

  def get_action_space(self, agent_id):
    return ActionSpace(ActionSpace.Continuous)

  def get_reward_min(self, agent_id):
    return 0

  def get_reward_max(self, agent_id):
    return 1

  def get_reward_fail(self, agent_id):
    return self.get_reward_min(agent_id)

  def get_reward_succ(self, agent_id):
    return self.get_reward_max(agent_id)

  #scene_name == "imitate" -> cDrawSceneImitate
  def get_state_size(self, agent_id):
    #cCtController::GetStateSize()
    #int state_size = cDeepMimicCharController::GetStateSize();
    #                     state_size += GetStatePoseSize();#106
    #                     state_size += GetStateVelSize(); #(3+3)*numBodyParts=90
    #state_size += GetStatePhaseSize();#1
    #197
    return 197

  def build_state_norm_groups(self, agent_id):
    #if (mEnablePhaseInput)
    #{
    #int phase_group = gNormGroupNone;
    #int phase_offset = GetStatePhaseOffset();
    #int phase_size = GetStatePhaseSize();
    #out_groups.segment(phase_offset, phase_size) = phase_group * Eigen::VectorXi::Ones(phase_size);
    groups = [0] * self.get_state_size(agent_id)
    groups[0] = -1
    return groups

  def build_state_offset(self, agent_id):
    out_offset = [0] * self.get_state_size(agent_id)
    phase_offset = -0.5
    out_offset[0] = phase_offset
    return np.array(out_offset)

  def build_state_scale(self, agent_id):
    out_scale = [1] * self.get_state_size(agent_id)
    phase_scale = 2
    out_scale[0] = phase_scale
    return np.array(out_scale)

  def get_goal_size(self, agent_id):
    return 0

  def get_action_size(self, agent_id):
    ctrl_size = 43  #numDof
    root_size = 7
    return ctrl_size - root_size

  def build_goal_norm_groups(self, agent_id):
    return np.array([])

  def build_goal_offset(self, agent_id):
    return np.array([])

  def build_goal_scale(self, agent_id):
    return np.array([])

  def build_action_offset(self, agent_id):
    out_offset = [0] * self.get_action_size(agent_id)
    out_offset = [
        0.0000000000, 0.0000000000, 0.0000000000, -0.200000000, 0.0000000000, 0.0000000000,
        0.0000000000, -0.200000000, 0.0000000000, 0.0000000000, 0.00000000, -0.2000000, 1.57000000,
        0.00000000, 0.00000000, 0.00000000, -0.2000000, 0.00000000, 0.00000000, 0.00000000,
        -0.2000000, -1.5700000, 0.00000000, 0.00000000, 0.00000000, -0.2000000, 1.57000000,
        0.00000000, 0.00000000, 0.00000000, -0.2000000, 0.00000000, 0.00000000, 0.00000000,
        -0.2000000, -1.5700000
    ]
    #see cCtCtrlUtil::BuildOffsetScalePDPrismatic and
    #see cCtCtrlUtil::BuildOffsetScalePDSpherical
    return np.array(out_offset)

  def build_action_scale(self, agent_id):
    out_scale = [1] * self.get_action_size(agent_id)
    #see cCtCtrlUtil::BuildOffsetScalePDPrismatic and
    #see cCtCtrlUtil::BuildOffsetScalePDSpherical
    out_scale = [
        0.20833333333333, 1.00000000000000, 1.00000000000000, 1.00000000000000, 0.25000000000000,
        1.00000000000000, 1.00000000000000, 1.00000000000000, 0.12077294685990, 1.00000000000000,
        1.000000000000, 1.000000000000, 0.159235668789, 0.159235668789, 1.000000000000,
        1.000000000000, 1.000000000000, 0.079617834394, 1.000000000000, 1.000000000000,
        1.000000000000, 0.159235668789, 0.120772946859, 1.000000000000, 1.000000000000,
        1.000000000000, 0.159235668789, 0.159235668789, 1.000000000000, 1.000000000000,
        1.000000000000, 0.107758620689, 1.000000000000, 1.000000000000, 1.000000000000,
        0.159235668789
    ]
    return np.array(out_scale)

  def build_action_bound_min(self, agent_id):
    #see cCtCtrlUtil::BuildBoundsPDSpherical
    out_scale = [-1] * self.get_action_size(agent_id)
    out_scale = [
        -4.79999999999, -1.00000000000, -1.00000000000, -1.00000000000, -4.00000000000,
        -1.00000000000, -1.00000000000, -1.00000000000, -7.77999999999, -1.00000000000,
        -1.000000000, -1.000000000, -7.850000000, -6.280000000, -1.000000000, -1.000000000,
        -1.000000000, -12.56000000, -1.000000000, -1.000000000, -1.000000000, -4.710000000,
        -7.779999999, -1.000000000, -1.000000000, -1.000000000, -7.850000000, -6.280000000,
        -1.000000000, -1.000000000, -1.000000000, -8.460000000, -1.000000000, -1.000000000,
        -1.000000000, -4.710000000
    ]

    return out_scale

  def build_action_bound_max(self, agent_id):
    out_scale = [1] * self.get_action_size(agent_id)
    out_scale = [
        4.799999999, 1.000000000, 1.000000000, 1.000000000, 4.000000000, 1.000000000, 1.000000000,
        1.000000000, 8.779999999, 1.000000000, 1.0000000, 1.0000000, 4.7100000, 6.2800000,
        1.0000000, 1.0000000, 1.0000000, 12.560000, 1.0000000, 1.0000000, 1.0000000, 7.8500000,
        8.7799999, 1.0000000, 1.0000000, 1.0000000, 4.7100000, 6.2800000, 1.0000000, 1.0000000,
        1.0000000, 10.100000, 1.0000000, 1.0000000, 1.0000000, 7.8500000
    ]
    return out_scale

  def set_mode(self, mode):
    self._mode = mode

  def need_new_action(self, agent_id):
    if self.t >= self.needs_update_time:
      self.needs_update_time = self.t + 1. / 30.
      return True
    return False

  def record_state(self, agent_id):
    state = self._humanoid.getState()

    return np.array(state)

  def record_goal(self, agent_id):
    return np.array([])

  def calc_reward(self, agent_id):
    kinPose = self._humanoid.computePose(self._humanoid._frameFraction)
    reward = self._humanoid.getReward(kinPose)
    return reward

  def set_action(self, agent_id, action):
    #print("action=",)
    #for a in action:
    #  print(a)
    #np.savetxt("pb_action.csv", action, delimiter=",")
    self.desiredPose = self._humanoid.convertActionToPose(action)
    #we need the target root positon and orientation to be zero, to be compatible with deep mimic
    self.desiredPose[0] = 0
    self.desiredPose[1] = 0
    self.desiredPose[2] = 0
    self.desiredPose[3] = 0
    self.desiredPose[4] = 0
    self.desiredPose[5] = 0
    self.desiredPose[6] = 0
    target_pose = np.array(self.desiredPose)

    #np.savetxt("pb_target_pose.csv", target_pose, delimiter=",")

    #print("set_action: desiredPose=", self.desiredPose)

  def log_val(self, agent_id, val):
    pass

  def update(self, timeStep):
    #print("pybullet_deep_mimic_env:update timeStep=",timeStep," t=",self.t)
    self._pybullet_client.setTimeStep(timeStep)
    self._humanoid._timeStep = timeStep
    self.timeStep = timeStep

    for i in range(1):
      self.t += timeStep
      self._humanoid.setSimTime(self.t)

      if self.desiredPose:
        kinPose = self._humanoid.computePose(self._humanoid._frameFraction)
        self._humanoid.initializePose(self._humanoid._poseInterpolator,
                                      self._humanoid._kin_model,
                                      initBase=True)
        #pos,orn=self._pybullet_client.getBasePositionAndOrientation(self._humanoid._sim_model)
        #self._pybullet_client.resetBasePositionAndOrientation(self._humanoid._kin_model, [pos[0]+3,pos[1],pos[2]],orn)
        #print("desiredPositions=",self.desiredPose)
        maxForces = [
            0, 0, 0, 0, 0, 0, 0, 200, 200, 200, 200, 50, 50, 50, 50, 200, 200, 200, 200, 150, 90,
            90, 90, 90, 100, 100, 100, 100, 60, 200, 200, 200, 200, 150, 90, 90, 90, 90, 100, 100,
            100, 100, 60
        ]

        if self._useStablePD:
          usePythonStablePD = False
          if usePythonStablePD:
            taus = self._humanoid.computePDForces(self.desiredPose,
                                                desiredVelocities=None,
                                                maxForces=maxForces)
            #taus = [0]*43
            self._humanoid.applyPDForces(taus)
          else:
            self._humanoid.computeAndApplyPDForces(self.desiredPose,
                                                maxForces=maxForces)
        else:
          self._humanoid.setJointMotors(self.desiredPose, maxForces=maxForces)

        self._pybullet_client.stepSimulation()

  def set_sample_count(self, count):
    return

  def check_terminate(self, agent_id):
    return Env.Terminate(self.is_episode_end())

  def is_episode_end(self):
    isEnded = self._humanoid.terminates()
    #also check maximum time, 20 seconds (todo get from file)
    #print("self.t=",self.t)
    if (self.t > 20):
      isEnded = True
    return isEnded

  def check_valid_episode(self):
    #could check if limbs exceed velocity threshold
    return True

  def getKeyboardEvents(self):
    return self._pybullet_client.getKeyboardEvents()

  def isKeyTriggered(self, keys, key):
    o = ord(key)
    #print("ord=",o)
    if o in keys:
      return keys[ord(key)] & self._pybullet_client.KEY_WAS_TRIGGERED
    return False



import pybullet as p
import time
import numpy as np

import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.loadURDF("plane.urdf")
p.loadURDF("sphere2.urdf",[0,0,2])

dt = 1./240.
p.setTimeStep(dt)

def getSceneAABB():
  aabbMins=[]
  aabbMaxs=[]

  for i in range(p.getNumBodies()):
    uid = p.getBodyUniqueId(i)
    aabb = p.getAABB(uid)
    aabbMins.append(np.array(aabb[0]))
    aabbMaxs.append(np.array(aabb[1]))

  if len(aabbMins):
    sceneAABBMin = aabbMins[0]
    sceneAABBMax = aabbMaxs[0]

    for aabb in aabbMins:
      sceneAABBMin = np.minimum(sceneAABBMin,aabb)
    for aabb in aabbMaxs:
      sceneAABBMax = np.maximum(sceneAABBMax,aabb)

    print("sceneAABBMin=",sceneAABBMin)
    print("sceneAABBMax=",sceneAABBMax)
 
getSceneAABB()

while (1):
  p.stepSimulation()
  time.sleep(dt)

